"""In a variable importance method, the ``scoring_strategy`` is a function which 
is used to determine which of the scores corresponding to a given variable 
indicates that the variable is "most important". This will be dependent on the
particular type of object which is returned as a score.

Here, we provide a few functions which can be used directly as scoring 
strategies as well as some utilities for construction scoring strategies. 
Moreover, we also provide a dictionary of aliases for several commonly used
strategies in ``VALID_SCORING_STRATEGIES``.
"""

import numpy as np

from .error_handling import InvalidStrategyException

__all__ = [
    "verify_scoring_strategy",
    "VALID_SCORING_STRATEGIES",
    "argmin_of_mean",
    "argmax_of_mean",
    "indexer_of_converter",
]


def verify_scoring_strategy(scoring_strategy):
    """Asserts that the scoring strategy is valid and interprets various strings

    :param scoring_strategy: a function to be used for determining optimal
        variables or a string. If a function, should be of the form
        ``([some value]) -> index``. If a string, must be one of the options in
        ``VALID_SCORING_STRATEGIES``
    :returns: a function to be used for determining optimal variables
    """
    if callable(scoring_strategy):
        return scoring_strategy
    elif scoring_strategy in VALID_SCORING_STRATEGIES:
        return VALID_SCORING_STRATEGIES[scoring_strategy]
    else:
        raise InvalidStrategyException(
            scoring_strategy, options=list(VALID_SCORING_STRATEGIES.keys())
        )


class indexer_of_converter(object):
    """This object is designed to help construct a scoring strategy by breaking
    the process of determining an optimal score into two pieces:
    First, each of the scores are converted to a simpler representation. For
    instance, an array of scores resulting from a bootstrapped evaluation method
    may be converted to just their mean.
    Second, each of the simpler representations are compared to determine the
    index of the one which is most optimal. This is typically just an ``argmin``
    or ``argmax`` call.
    """

    def __init__(self, indexer, converter):
        """Constructs a function which first converts all objects in a list to
        something simpler and then uses the indexer to determine the index of
        the most "optimal" one

        :param indexer: a function which converts a list of probably simply
            values (like numbers) to a single index
        :param converter: a function which converts a single more complex object
            to a simpler one (like a single number)
        """
        self.indexer = indexer
        self.converter = converter

    def __call__(self, scores):
        """Finds the index of the most "optimal" score in a list"""
        return self.indexer([self.converter(score) for score in scores])


argmin_of_mean = indexer_of_converter(np.argmin, np.mean)
argmax_of_mean = indexer_of_converter(np.argmax, np.mean)


VALID_SCORING_STRATEGIES = {
    "max": argmax_of_mean,
    "maximize": argmax_of_mean,
    "argmax": np.argmax,
    "min": argmin_of_mean,
    "minimize": argmin_of_mean,
    "argmin": np.argmin,
    "argmin_of_mean": argmin_of_mean,
    "argmax_of_mean": argmax_of_mean,
}



import xarray as xr
import numpy as np
from skexplain.common.utils import compute_bootstrap_indices
import pandas as pd

def method_average_ranking(data, features, methods, estimator_names, n_features=12):
    """
    Compute the median ranking across the results of different ranking methods.
    Also, include the 25-75th percentile ranking uncertainty.

    Parameters
    ------------
        data : list of xarray.Dataset
            The set of predictor ranking results to average over.

        methods : list of string
            The ranking methods to use from the data (see plot_importance for examples)

        estimator_names : string or list of strings
            Name of the estimator(s).

    Returns
    --------
        rankings_dict_avg : dict
            feature : median ranking pairs
        rankings_sorted : np.array
            Sorted median rankings (lower values indicates higher rank)
        feature_sorted : np.array
            The features corresponding the ranks in ``rankings_sorted``
        xerr

    """
    rankings_dict = {f: [] for f in features}
    for d, method in zip(data, methods):
        for estimator_name in estimator_names:
            features = d[f"{method}_rankings__{estimator_name}"].values[:n_features]
            rankings = {f: i for i, f in enumerate(features)}
            for f in features:
                try:
                    rankings_dict[f].append(rankings[f])
                except:
                    rankings_dict[f].append(np.nan)

    max_len = np.max([len(rankings_dict[k]) for k in rankings_dict.keys()])
    for k in rankings_dict.keys():
        l = rankings_dict[k]
        if len(l) < max_len:
            delta = max_len - len(l)
            rankings_dict[k] = l + [np.nan] * delta

    rankings_dict_avg = {
        f: np.nanpercentile(rankings_dict[f], 50) for f in rankings_dict.keys()
    }

    features = np.array(list(rankings_dict_avg.keys()))
    rankings = np.array([rankings_dict_avg[f] for f in features])
    idxs = np.argsort(rankings)

    rankings_sorted = rankings[idxs]
    features_ranked = features[idxs]

    scores = np.array([rankings_dict[f] for f in features_ranked])

    data = {}
    data[f"combined_rankings__{estimator_name}"] = (
        [f"n_vars_avg"],
        features_ranked,
    )
    data[f"combined_scores__{estimator_name}"] = (
        [f"n_vars_avg", "n_bootstrap"],
        scores,
    )
    data = xr.Dataset(data)

    return data


def non_increasing(L):
    # Check for decreasing scores.
    return all(x >= y for x, y in zip(L, L[1:]))


def compute_importance(results, scoring_strategy, direction):
    """
    Compute the importance scores from the permutation importance results.
    The importance score varies depending on the orientation of the
    loss metric and whether it is multipass or singlepass.


    Parameters
    --------------
    results : InterpretToolkit.permutation_importance results
              xr.Dataset
    scoring_strategy : 'minimize' or 'maximize'
        Whether the strategy for assessing importance was 
        based on minimizing or maximing the performance metric 
        after permuting features (e.g., the goal is to 'maximize' 
        loss metrics like MSE, but 'minimize' rank metrics like AUC)

    direction : 'forward' or 'backward'
        Whether the permutation method was 'forward' or 'backward'.


    Returns
    --------------
    results : xarray.Dataset
        scores for each estimator and multi/singlepass are
        converted to proper importance scores.
    """
    if scoring_strategy == 'argmin_of_mean':
        scoring_strategy = 'minimize'
    elif scoring_strategy == 'argmax_of_mean':
        scoring_strategy = 'maximize' 
    
    print(direction, scoring_strategy)
    estimators = results.attrs["estimators used"]
    for estimator in estimators:
        orig_score = results[f"original_score__{estimator}"].values
        if direction == 'forward':
            all_permuted_score = results[f"all_permuted_score__{estimator}"].values
            
        for mode in ["singlepass", "multipass"]:
            permute_scores = results[f"{mode}_scores__{estimator}"].values

            if direction == 'forward':
                # For a loss metric, forward importance is generically defined 
                # as the error(X_J') - error(X_j') where J is the set of 
                # all features while j is some subset of J or a single feature.
                if scoring_strategy == 'maximize':
                    # E.g., AUC, NAUPDC, CSI, BSS
                    imp = permute_scores - all_permuted_score
                   
                elif scoring_strategy == 'minimize':
                # For a rank-based metric, importance is defined opposite 
                # of the loss metric [ error(X_j') - error(X_J') ] 
                    # E.g., MSE, BS, etc.
                    print('This happened for forward!')
                    imp = all_permuted_score - permute_scores
                    
            elif direction == 'backward':
                #  For a loss metric, backward importance is generically defined
                # as the error(X_j') - error(X_j) where j is some subset or
                # a single feature. 
                if scoring_strategy == 'minimize':
                    imp = orig_score - permute_scores
                elif scoring_strategy == 'maximize':
                # For a rank-based metric, it is defined opposite of that above.
                # i.e., error(X_j) - error(X_j') 
                    
                    print('this happened for backward!')
                    imp = permute_scores - orig_score

            
            """
            decreasing = non_increasing(np.mean(permute_scores, axis=1))

            if decreasing:
                if orientation == "negative":
                    # Singlepass MSE
                    imp = permute_scores - orig_score
                else:
                    # Backward Multipass on AUC/AUPDC (permuted_score - (1-orig_score)).
                    # Most positively-oriented metrics top off at 1.
                    top = np.max(permute_scores)
                    imp = permute_scores - (top - orig_score)
            else:
                if orientation == "negative":
                    # Forward Multipass MSE
                    top = np.max(permute_scores)
                    imp = (top + orig_score) - permute_scores
                else:
                    # Singlepass AUC/NAUPDC
                    imp = orig_score - permute_scores
            """

            # Normalize the importance score so that range is [0,1]
            imp = imp / (np.percentile(imp, 99) - np.percentile(imp, 1))

            results[f"{mode}_scores__{estimator}"] = (
                [f"n_vars_{mode}", "n_bootstrap"],
                imp,
            )

    return results


def to_skexplain_importance(
    importances, estimator_name, feature_names, method, normalize=True
):
    """
    Convert the feature ranking-based scores from non-permutation-importance methods
    into a importance dataset for plotting purposes
    """

    bootstrap = False
    if method == "sage":
        importances_std = importances.std
        importances = importances.values
    elif method == "coefs":
        importances = np.absolute(importances)
    elif method == "shap_std":
        # Compute the std(SHAP)
        importances = np.std(importances, axis=0)
    elif method == "shap_sum":
        # Compute sum of abs values
        importances = np.sum(np.absolute(importances), axis=0)
    else:
        if np.ndim(importances) == 2:
            # average over bootstrapping
            bootstrap = True
            importances_to_save = importances.copy()
            importances = np.mean(importances, axis=1)

    # Sort from higher score to lower score
    ranked_indices = np.argsort(importances)[::-1]

    if bootstrap:
        scores_ranked = importances_to_save[ranked_indices, :]
    else:
        scores_ranked = importances[ranked_indices]

    if method == "sage":
        std_ranked = importances_std[ranked_indices]

    features_ranked = np.array(feature_names)[ranked_indices]

    data = {}
    data[f"{method}_rankings__{estimator_name}"] = (
        [f"n_vars_{method}"],
        features_ranked,
    )

    if not bootstrap:
        scores_ranked = scores_ranked.reshape(len(scores_ranked), 1)
        importances = importances.reshape(len(importances), 1)

    if normalize:
        # Normalize the importance score so that range is [0,1]
        scores_ranked = scores_ranked / (
            np.percentile(scores_ranked, 99) - np.percentile(scores_ranked, 1)
        )

    data[f"{method}_scores__{estimator_name}"] = (
        [f"n_vars_{method}", "n_bootstrap"],
        scores_ranked,
    )

    if method == "sage":
        data[f"sage_scores_std__{estimator_name}"] = (
            [f"n_vars_sage"],
            std_ranked,
        )

    data = xr.Dataset(data)

    data.attrs["estimators used"] = estimator_name
    data.attrs["estimator output"] = "probability"

    return data


def combine_top_features(results_dict, n_vars=None):
    """Combines the list of top features from different estimators
    into a single list where duplicates are removed.

    Args:
    -------------
        results_dict : dict
        n_vars : integer
    """
    if n_vars is None:
        n_vars = 1000
    combined_features = []
    for estimator_name in results_dict.keys():
        features = results_dict[estimator_name]
        combined_features.append(features)
    unique_features = list(set.intersection(*map(set, combined_features)))[:n_vars]

    return unique_features


def retrieve_important_vars(results, estimator_names, multipass=True):
    """
    Return a list of the important features stored in the
     ImportanceObject

     Args:
     -------------------
         results : python object
             ImportanceObject from PermutationImportance
         multipass : boolean
             if True, returns the multipass permutation importance results
             else returns the singlepass permutation importance results

     Returns:
         top_features : list
             a list of features with order determined by
             the permutation importance method
    """
    perm_method = "multipass" if multipass else "singlepass"

    important_vars_dict = {}
    for estimator_name in estimator_names:
        top_features = list(results[f"{perm_method}_rankings__{estimator_name}"].values)
        important_vars_dict[estimator_name] = top_features

    return important_vars_dict


def find_correlated_pairs_among_top_features(
    corr_matrix,
    top_features,
    rho_threshold=0.8,
):
    """
    Of the top features, find correlated pairs above some
    linear correlation coefficient threshold

    Args:
    ----------------------
        corr_matrix : pandas.DataFrame
        top_features : list of strings
        rho_threshold : float

    """
    top_feature_indices = {f: i for i, f in enumerate(top_features)}
    sub_corr_matrix = corr_matrix[top_features].loc[top_features]

    pairs = []
    for feature in top_features:
        # try:
        most_corr_feature = (
            sub_corr_matrix[feature].sort_values(ascending=False).index[1]
        )
        # except:
        #    continue

        most_corr_value = sub_corr_matrix[feature].sort_values(ascending=False)[1]
        if round(most_corr_value, 5) >= rho_threshold:
            pairs.append((feature, most_corr_feature))

    pairs = list(set([tuple(sorted(t)) for t in pairs]))
    pair_indices = [
        (top_feature_indices[p[0]], top_feature_indices[p[1]]) for p in pairs
    ]

    return pairs, pair_indices

def all_permuted_score(estimator, X, y, evaluation_fn, n_permute, subsample, random_seed=123):
    random_state = np.random.RandomState(random_seed)
    inds = random_state.permutation(len(X))
    if isinstance(X, pd.DataFrame):
        X = X.values
    
    inds_set = compute_bootstrap_indices(X, subsample=1.0, n_bootstrap=n_permute, seed=90)
    
    scores = []
    
    for inds in inds_set:
        X_sampled = X[inds, :]
        
        X_permuted = np.array([ X_sampled[inds, i] for i in range(X.shape[1])]).T
    
        if hasattr(estimator, 'predict_proba'):
            predictions = estimator.predict_proba(X_permuted)[:,1]
        elif hasattr(estimator, 'predict'):
            predictions = estimator.predict(X_permuted)[:]
        else:
            raise AttributeError(f'{estimator} does not have .predict or .predict_proba!')
    
        scores.append(evaluation_fn(y, predictions))
        
    return np.array(scores)
    




"""The general algorithm for all of the data-based variable importance methods
is the same, regardless of whether the method is Sequential Selection or 
Permutation Importance or something else. This is represented in the 
``abstract_variable_importance`` function. All of the different methods we 
provide use this function under the hood and the only difference between them is
the ``selection_strategy`` object, which is detailed in 
:mod:`PermutationImportance.selection_strategies`. Typically, you will not need 
to use this method but can instead use one of the methods imported directly into 
the top package of **PermutationImportance**.

If you wish to implement your own variable importance method, you will need to
devise your own ``selection_strategy``. We recommend using
:mod:`PermutationImportance.selection_strategies` as a template for implementing 
your own variable importance method."""

import numpy as np
import multiprocessing as mp

from .data_verification import verify_data, determine_variable_names
from .multiprocessing_utils import pool_imap_unordered
from .result import ImportanceResult
from .scoring_strategies import verify_scoring_strategy
from .utils import add_ranks_to_dict, get_data_subset, bootstrap_generator

from ...common.multiprocessing_utils import run_parallel
from ...common.utils import merge_dict

def abstract_variable_importance(
    training_data,
    scoring_data,
    scoring_fn,
    scoring_strategy,
    selection_strategy,
    variable_names=None,
    nimportant_vars=None,
    method=None,
    perm_method="forwards",
    njobs=1,
    verbose=False,
    random_seed=1,
    **kwargs,
):
    """Performs an abstract variable importance over data given a particular
    set of functions for scoring, determining optimal variables, and selecting
    data

    :param training_data: a 2-tuple ``(inputs, outputs)`` for training in the
        ``scoring_fn``
    :param scoring_data: a 2-tuple ``(inputs, outputs)`` for scoring in the
        ``scoring_fn``
    :param scoring_fn: a function to be used for scoring. Should be of the form
        ``(training_data, scoring_data) -> some_value``
    :param scoring_strategy: a function to be used for determining optimal
        variables. Should be of the form ``([some_value]) -> index``
    :param variable_names: an optional list for variable names. If not given,
        will use names of columns of data (if pandas dataframe) or column
        indices
    :param nimportant_vars: number of variables to compute importance for.
        Defaults to all variables
    :param method: a string for the name of the method used. Defaults to the
        name of the ``selection_strategy`` if not given
    :param njobs: an integer for the number of threads to use. If negative, will
        use ``num_cpus + njobs``. Defaults to 1
    :returns: :class:`PermutationImportance.result.ImportanceResult` object
        which contains the results for each run
    """
    #training_data = #verify_data(training_data)
    scoring_data = verify_data(scoring_data)
    
    # Sending the scoring_data in as training_data so that for the 
    # forward permutation importance methods, we can un-permuted data. 
    training_data = (scoring_data[0].copy(), scoring_data[1].copy())
    
    scoring_strategy = verify_scoring_strategy(scoring_strategy)
    variable_names = determine_variable_names(scoring_data, variable_names)

    nimportant_vars = (
        len(variable_names) if nimportant_vars is None else nimportant_vars
    )

    method = (
        getattr(selection_strategy, "name", getattr(selection_strategy, "__name__"))
        if method is None
        else method
    )

    njobs = mp.cpu_count() if njobs <= 0 else njobs
    njobs = int(njobs * mp.cpu_count()) if (njobs < 1 and njobs >= 0) else njobs

    important_vars = list()
    num_vars = len(variable_names)

    # This random state generator is for the predictors left permuted.
    # As predictors are left permuted, they are left different permuted states
    # with each multi-pass iterations. This hopefully ensures that the permuted
    # variables are not left in poor permutations to bias the results.
    random_states = bootstrap_generator(n_bootstrap=nimportant_vars, seed=156)
    permute_rs = bootstrap_generator(n_bootstrap=scoring_fn.n_permute, seed=42)
    
    subsample_size = scoring_fn.get_subsample_size(full_size=len(scoring_data[1]))
    
    rows = [rs.choice(len(scoring_data[1]), subsample_size) for rs in permute_rs]
    inds = [rs.permutation(subsample_size) for rs in permute_rs]
   
    # Setting the shuffled indices to reduce computational costs. 
    scoring_fn.shuffled_indices = inds
    scoring_fn.rows = rows 

    # Compute the original score over all the data
    original_score = scoring_fn(training_data, scoring_data, var_idx=None)
    result_obj = ImportanceResult(method, variable_names, original_score)
    
    for i, _ in enumerate(range(nimportant_vars)):
        if verbose:
            print(f"Multi-pass iteration {i+1} out of {nimportant_vars}...")

        # Sending the scoring_data in as training_data so that for the 
        # forward permutation importance methods, we can un-permuted data. 
        selection_iter = selection_strategy(
            training_data,
            scoring_data,
            num_vars,
            important_vars,
            random_states[i],
            **kwargs,
        )

        if njobs == 1:
            result = _singlethread_iteration(selection_iter, scoring_fn)
        else:
            result = _multithread_iteration(selection_iter, scoring_fn, njobs, nimportant_vars-i)
   
        next_result = add_ranks_to_dict(result, variable_names, scoring_strategy)
        best_var = min(next_result.keys(), key=lambda key: next_result[key][0])
        best_index = np.flatnonzero(variable_names == best_var)[0]
        result_obj.add_new_results(next_result, next_important_variable=best_var)
        important_vars.append(best_index)

    return result_obj


def _singlethread_iteration(selection_iterator, scoring_fn):
    """Handles a single pass of the abstract variable importance algorithm,
    assuming a single worker thread

    :param selection_iterator: an iterator which yields triples
        ``(variable, training_data, scoring_data)``. Typically a
        :class:`PermutationImportance.selection_strategies.SelectionStrategy`
    :param scoring_fn: a function to be used for scoring. Should be of the form
        ``(training_data, scoring_data) -> float``
    :returns: a dict of ``{var: score}``
    """
    result = dict()
    for training_data, scoring_data, var in selection_iterator:
        score = scoring_fn(training_data, scoring_data, var_idx=var)
        result[var] = score
    return result


def _multithread_iteration(selection_iterator, scoring_fn, njobs, total=None):
    """Handles a single pass of the abstract variable importance algorithm using
    multithreading

    :param selection_iterator: an iterator which yields triples
        ``(variable, training_data, scoring_data)``. Typically a
        :class:`PermutationImportance.selection_strategies.SelectionStrategy`
    :param scoring_fn: a function to be used for scoring. Should be of the form
        ``(training_data, scoring_data) -> float``
    :param num_jobs: number of processes to use
    :returns: a dict of ``{var: score}``
    """
    result = dict()
    for index, score in pool_imap_unordered(scoring_fn, selection_iterator, njobs):
        result[index] = score
    return result

    """
    This appears to be faster for cases when the dataset is smaller (<5-10K)
    def worker(training_data, scoring_data, var):
        result = {}
        score = scoring_fn(training_data, scoring_data, var_idx=var)
        result[var] = score

        return result 
    
    result = run_parallel(func=worker,
                args_iterator=selection_iterator,
                kwargs={},
                nprocs_to_use=njobs,
                total=total,
                         )
    return merge_dict(result)
    """




# import matplotlib
# matplotlib.use('Agg')

import numpy as np

import matplotlib.pyplot as plt
from matplotlib.ticker import MaxNLocator, FormatStrFormatter, AutoMinorLocator
import matplotlib.ticker as mticker

from matplotlib import rcParams
from matplotlib.colors import ListedColormap
from matplotlib.gridspec import GridSpec
import matplotlib
import seaborn as sns

from ..common.utils import is_outlier
from ..common.contrib_utils import combine_like_features
import shap

class PlotStructure:
    """
    Plot handles figure and subplot generation
    """

    def __init__(self, BASE_FONT_SIZE=12, set_seaborn=True):
        
        GENERIC_FONT_SIZE_NAMES = [
            "teensie",
            "tiny",
            "small",
            "normal",
            "big",
            "large",
            "huge",
            ]

        FONT_SIZES_ARRAY = np.arange(-6, 8, 2) + BASE_FONT_SIZE

        self.FONT_SIZES = {
                name: size for name, size in zip(GENERIC_FONT_SIZE_NAMES, FONT_SIZES_ARRAY)
            }
        
        if set_seaborn:
            sns.set_theme()
        
        # Setting the font style to serif
        rcParams["font.family"] = "serif"

        plt.rc("font", size=self.FONT_SIZES["normal"])  # controls default text sizes
        plt.rc("axes", titlesize=self.FONT_SIZES["tiny"])  # fontsize of the axes title
        plt.rc(
                 "axes", labelsize=self.FONT_SIZES["normal"]
                )  # fontsize of the x and y labels
        plt.rc(
                "xtick", labelsize=self.FONT_SIZES["teensie"]
                )  # fontsize of the x-axis tick marks
        plt.rc(
                "ytick", labelsize=self.FONT_SIZES["teensie"]
                )  # fontsize of the y-axis tick marks
        plt.rc("legend", fontsize=self.FONT_SIZES["teensie"])  # legend fontsize
        plt.rc(
                "figure", titlesize=self.FONT_SIZES["big"]
                )  # fontsize of the figure title

    def get_fig_props(self, n_panels, **kwargs):
        """Determine appropriate figure properties"""
        width_slope = 0.875
        height_slope = 0.45
        intercept = 3.0 - width_slope
        figsize = (
            min((n_panels * width_slope) + intercept, 19),
            min((n_panels * height_slope) + intercept, 12),
        )

        wspace = (-0.03 * n_panels) + 0.85
        hspace = (0.0175 * n_panels) + 0.3

        n_columns = kwargs.get("n_columns", 3)
        wspace = wspace + 0.25 if n_columns > 3 else wspace

        kwargs["figsize"] = kwargs.get("figsize", figsize)
        kwargs["wspace"] = kwargs.get("wspace", wspace)
        kwargs["hspace"] = kwargs.get("hspace", hspace)

        return kwargs

    def create_subplots(self, n_panels, **kwargs):
        """
        Create a series of subplots (MxN) based on the
        number of panels and number of columns (optionally).

        Args:
        -----------------------
            n_panels : int
                Number of subplots to create
            Optional keyword args:
                n_columns : int
                    The number of columns for a plot (default=3 for n_panels >=3)
                figsize: 2-tuple of figure size (width, height in inches)
                wspace : float
                    the amount of width reserved for space between subplots,
                    expressed as a fraction of the average axis width
                hspace : float
                sharex : boolean
                sharey : boolean
        """
        # figsize = width, height in inches

        figsize = kwargs.get("figsize", (6.4, 4.8))
        wspace = kwargs.get("wspace", 0.4)
        hspace = kwargs.get("hspace", 0.3)
        sharex = kwargs.get("sharex", False)
        sharey = kwargs.get("sharey", False)

        delete = True
        if n_panels <= 3:
            n_columns = kwargs.get("n_columns", n_panels)
            delete = True if n_panels != n_columns else False
        else:
            n_columns = kwargs.get("n_columns", 3)

        n_rows = int(n_panels / n_columns)
        extra_row = 0 if (n_panels % n_columns) == 0 else 1

        fig, axes = plt.subplots(
            n_rows + extra_row,
            n_columns,
            sharex=sharex,
            sharey=sharey,
            figsize=figsize,
            dpi=300,
        )

        fig.patch.set_facecolor("white")

        plt.subplots_adjust(wspace=wspace, hspace=hspace)

        if delete:
            n_axes_to_delete = len(axes.flat) - n_panels

            if n_axes_to_delete > 0:
                for i in range(n_axes_to_delete):
                    fig.delaxes(axes.flat[-(i + 1)])

        return fig, axes

    def _create_joint_subplots(self, n_panels, **kwargs):
        """
        Create grid for multipanel drawing a bivariate plots with marginal
        univariate plots on the top and right hand side.
        """
        figsize = kwargs.get("figsize", (6.4, 4.8))
        ratio = kwargs.get("ratio", 5)
        n_columns = kwargs.get("n_columns", 3)

        fig = plt.figure(figsize=figsize, dpi=300)
        fig.patch.set_facecolor("white")
        extra_row = 0 if (n_panels % n_columns) == 0 else 1

        nrows = ratio * (int(n_panels / n_columns) + extra_row)
        ncols = ratio * n_columns

        gs = GridSpec(ncols=ncols, nrows=nrows)

        main_ax_len = ratio - 1

        main_axes = []
        top_axes = []
        rhs_axes = []

        col_offset_idx = list(range(n_columns)) * int(nrows / ratio)

        row_offset = 0
        for i in range(n_panels):
            col_offset = ratio * col_offset_idx[i]

            row_idx = 1
            if i % n_columns == 0 and i > 0:
                row_offset += ratio

            main_ax = fig.add_subplot(
                gs[
                    row_idx + row_offset : main_ax_len + row_offset,
                    col_offset : main_ax_len + col_offset - 1,
                ],
                frameon=False,
            )

            top_ax = fig.add_subplot(
                gs[row_idx + row_offset - 1, col_offset : main_ax_len + col_offset - 1],
                frameon=False,
                sharex=main_ax,
            )

            rhs_ax = fig.add_subplot(
                gs[
                    row_idx + row_offset : main_ax_len + row_offset,
                    main_ax_len + col_offset - 1,
                ],
                frameon=False,
                sharey=main_ax,
            )

            ax_marg = [top_ax, rhs_ax]
            for ax in ax_marg:
                # Turn off tick visibility for the measure axis on the marginal plots
                plt.setp(ax.get_xticklabels(), visible=False)
                plt.setp(ax.get_yticklabels(), visible=False)
                # Turn off the ticks on the density axis for the marginal plots
                plt.setp(ax.yaxis.get_majorticklines(), visible=False)
                plt.setp(ax.xaxis.get_majorticklines(), visible=False)
                plt.setp(ax.yaxis.get_minorticklines(), visible=False)
                plt.setp(ax.xaxis.get_minorticklines(), visible=False)
                ax.yaxis.grid(False)
                ax.xaxis.grid(False)
                for axis in [ax.xaxis, ax.yaxis]:
                    axis.label.set_visible(False)

            main_axes.append(main_ax)
            top_axes.append(top_ax)
            rhs_axes.append(rhs_ax)

        n_rows = int(nrows / ratio)

        return fig, main_axes, top_axes, rhs_axes, n_rows

    def axes_to_iterator(self, n_panels, axes):
        """Turns axes list into iterable"""
        if isinstance(axes, list):
            return axes
        else:
            ax_iterator = [axes] if n_panels == 1 else axes.flat
            return ax_iterator

    def set_major_axis_labels(
        self,
        fig,
        xlabel=None,
        ylabel_left=None,
        ylabel_right=None,
        title=None,
        **kwargs,
    ):
        """
        Generate a single X- and Y-axis labels for
        a series of subplot panels. E.g.,
        """
        fontsize = kwargs.get("fontsize", self.FONT_SIZES["normal"])
        labelpad = kwargs.get("labelpad", 15)
        ylabel_right_color = kwargs.get("ylabel_right_color", "k")

        # add a big axis, hide frame
        ax = fig.add_subplot(111, frameon=False)

        # hide tick and tick label of the big axis
        plt.tick_params(
            labelcolor="none", top=False, bottom=False, left=False, right=False
        )

        # set axes labels
        ax.set_xlabel(xlabel, fontsize=fontsize, labelpad=labelpad)
        ax.set_ylabel(ylabel_left, fontsize=fontsize, labelpad=labelpad)

        if ylabel_right is not None:
            ax_right = fig.add_subplot(1, 1, 1, sharex=ax, frameon=False)
            plt.tick_params(
                labelcolor="none", top=False, bottom=False, left=False, right=False
            )

            ax_right.yaxis.set_label_position("right")
            ax_right.set_ylabel(
                ylabel_right,
                labelpad=2 * labelpad,
                fontsize=fontsize,
                color=ylabel_right_color,
            )
            ax_right.grid(False)

        ax.set_title(title)
        
        ax.grid(False)

        return ax

    def set_row_labels(self, labels, axes, pos=-1, pad=1.15, rotation=270, **kwargs):
        """
        Give a label to each row in a series of subplots
        """
        colors = kwargs.get("colors", ["xkcd:darkish blue"] * len(labels))
        fontsize = kwargs.get("fontsize", self.FONT_SIZES["small"])

        if np.ndim(axes) == 2:
            iterator = axes[:, pos]
        else:
            iterator = [axes[pos]]

        for ax, row, color in zip(iterator, labels, colors):
            ax.yaxis.set_label_position("right")
            ax.annotate(
                row,
                xy=(1, 1),
                xytext=(pad, 0.5),
                xycoords=ax.transAxes,
                rotation=rotation,
                size=fontsize,
                ha="center",
                va="center",
                color=color,
                alpha=0.65,
            )

    def add_alphabet_label(self, n_panels, axes, pos=(0.9, 0.09), alphabet_fontsize=10, **kwargs):
        """
        A alphabet character to each subpanel.
        """
        alphabet_list = [chr(x) for x in range(ord("a"), ord("z") + 1)] + [
            f"{chr(x)}{chr(x)}" for x in range(ord("a"), ord("z") + 1)
        ]

        ax_iterator = self.axes_to_iterator(n_panels, axes)

        for i, ax in enumerate(ax_iterator):
            ax.text(
                pos[0],
                pos[1],
                f"({alphabet_list[i]})",
                fontsize=alphabet_fontsize,
                alpha=0.8,
                ha="center",
                va="center",
                transform=ax.transAxes,
            )

    def _to_sci_notation(self, ydata, ax=None, xdata=None, colorbar=False):
        """
        Convert decimals (less 0.01) to 10^e notation
        """
        # f = mticker.ScalarFormatter(useOffset=False, useMathText=True)
        # g = lambda x, pos: "${}$".format(f._formatSciNotation("%1.10e" % x))

        if colorbar and np.absolute(np.amax(ydata)) <= 0.01:
            # colorbar.ax.yaxis.set_major_formatter(mticker.FuncFormatter(g))
            colorbar.ax.ticklabel_format(
                style="sci",
            )
            colorbar.ax.tick_params(axis="y", labelsize=5)
        elif ax:
            if np.absolute(np.amax(xdata)) <= 0.01:
                ax.ticklabel_format(
                    style="sci",
                )
                # ax.xaxis.set_major_formatter(mticker.FuncFormatter(g))
                ax.tick_params(axis="x", labelsize=5, rotation=45)
            if np.absolute(np.amax(ydata)) <= 0.01:
                # ax.yaxis.set_major_formatter(mticker.FuncFormatter(g))
                ax.ticklabel_format(
                    style="sci",
                )
                ax.tick_params(axis="y", labelsize=5, rotation=45)

    def calculate_ticks(
        self,
        nticks,
        ax=None,
        upperbound=None,
        lowerbound=None,
        round_to=5,
        center=False,
    ):
        """
        Calculate the y-axis ticks marks for the line plots
        """
        if ax is not None:
            upperbound = round(ax.get_ybound()[1], 5)
            lowerbound = round(ax.get_ybound()[0], 5)

        max_value = max(abs(upperbound), abs(lowerbound))
        if 0 < max_value < 1:
            if max_value < 0.1:
                round_to = 3
            else:
                round_to = 5

        elif 5 < max_value < 10:
            round_to = 2
        else:
            round_to = 0

        def round_to_a_base(a_number, base=5):
            return base * round(a_number / base)

        if max_value > 5:
            max_value = round_to_a_base(max_value)

        if center:
            values = np.linspace(-max_value, max_value, nticks)
            values = np.round(values, round_to)
        else:
            dy = upperbound - lowerbound
            # deprecated 8 March 2022 by Monte.
            if round_to > 2:
                fit = np.floor(dy / (nticks - 1)) + 1
                dy = (nticks - 1) * fit
            values = np.linspace(lowerbound, lowerbound + dy, nticks)
            values = np.round(values, round_to)

        return values

    def set_tick_labels(
        self, ax, feature_names, display_feature_names, return_labels=False
    ):
        """
        Setting the tick labels for the tree interpreter plots.
        """
        if isinstance(display_feature_names, dict):
            labels = [
                display_feature_names.get(feature_name, feature_name)
                for feature_name in feature_names
            ]
        else:
            labels = display_feature_names

        if return_labels:
            labels = [f"{l}" for l in labels]
            return labels
        else:
            labels = [f"{l}" for l in labels]
            ax.set_yticklabels(labels)

    def set_axis_label(self, ax, xaxis_label=None, yaxis_label=None, **kwargs):
        """
        Setting the x- and y-axis labels with fancy labels (and optionally
        physical units)
        """
        fontsize = kwargs.get("fontsize", self.FONT_SIZES["tiny"])
        if xaxis_label is not None:
            xaxis_label_pretty = self.display_feature_names.get(
                xaxis_label, xaxis_label
            )
            units = self.display_units.get(xaxis_label, "")
            if units == "":
                xaxis_label_with_units = f"{xaxis_label_pretty}"
            else:
                xaxis_label_with_units = f"{xaxis_label_pretty} ({units})"

            ax.set_xlabel(xaxis_label_with_units, fontsize=fontsize)

        if yaxis_label is not None:
            yaxis_label_pretty = self.display_feature_names.get(
                yaxis_label, yaxis_label
            )
            units = self.display_units.get(yaxis_label, "")
            if units == "":
                yaxis_label_with_units = f"{yaxis_label_pretty}"
            else:
                yaxis_label_with_units = f"{yaxis_label_pretty} ({units})"

            ax.set_ylabel(yaxis_label_with_units, fontsize=fontsize)

    def set_legend(self, n_panels, fig, ax, major_ax, **kwargs):
        """
        Set a single legend on the bottom of a figure
        for a set of subplots.
        """
        fontsize = kwargs.get("fontsize", "medium")
        ncol = kwargs.get("ncol", 3)
        handles = kwargs.get("handles", None)
        labels = kwargs.get("labels", None)
        if handles is None and labels is None:
            handles, labels = ax.get_legend_handles_labels()

        if n_panels > 3:
            bbox_to_anchor = (0.5, -0.35)
        else:
            bbox_to_anchor = (0.5, -0.5)

        bbox_to_anchor = kwargs.get("bbox_to_anchor", bbox_to_anchor)

        # Shrink current axis's height by 10% on the bottom
        box = major_ax.get_position()
        major_ax.set_position(
            [box.x0, box.y0 + box.height * 0.1, box.width, box.height * 0.9]
        )

        # Put a legend below current axis
        major_ax.legend(
            handles,
            labels,
            loc="lower center",
            bbox_to_anchor=bbox_to_anchor,
            fancybox=True,
            shadow=True,
            ncol=ncol,
            fontsize=fontsize,
        )

    def set_minor_ticks(self, ax):
        """
        Adds minor tick marks to the x- and y-axis to a subplot ax
        to increase readability.
        """
        ax.xaxis.set_minor_locator(AutoMinorLocator(n=3))
        ax.yaxis.set_minor_locator(AutoMinorLocator(n=3))

    def set_n_ticks(self, ax, option="y", nticks=5):
        """
        Set the max number of ticks per x- and y-axis for a
        subplot ax
        """
        if option == "y" or option == "both":
            ax.yaxis.set_major_locator(MaxNLocator(nticks))
            
        if option == "x" or option == "both":
            ax.xaxis.set_major_locator(MaxNLocator(nticks))

    def make_twin_ax(self, ax):
        """
        Create a twin axis on an existing axis with a shared x-axis
        """
        # align the twinx axis
        twin_ax = ax.twinx()

        # Turn twin_ax grid off.
        twin_ax.grid(False)

        # Set ax's patch invisible
        ax.patch.set_visible(False)
        # Set axtwin's patch visible and colorize it in grey
        twin_ax.patch.set_visible(True)

        # move ax in front
        ax.set_zorder(twin_ax.get_zorder() + 1)

        return twin_ax

    def despine_plt(self, ax):
        """
        remove all four spines of plot
        """
        ax.spines["right"].set_visible(False)
        ax.spines["top"].set_visible(False)
        ax.spines["left"].set_visible(False)
        ax.spines["bottom"].set_visible(False)

    def annotate_bars(self, ax, num1, num2, y, width, dh=0.01, barh=0.05, delta=0):
        """ 
        Annotate barplot with connections between correlated variables. 

        Parameters
        ----------------
            num1: index of left bar to put bracket over
            num2: index of right bar to put bracket over
            y: centers of all bars (like plt.barh() input)
            width: widths of all bars (like plt.barh() input)
            dh: height offset over bar / bar + yerr in axes coordinates (0 to 1)
            barh: bar height in axes coordinates (0 to 1)
            delta : shifting of the annotation when multiple annotations would overlap
        """
        lx, ly = y[num1], width[num1]
        rx, ry = y[num2], width[num2]

        ax_y0, ax_y1 = plt.gca().get_xlim()
        dh *= (ax_y1 - ax_y0)
        barh *= (ax_y1 - ax_y0)

        y = max(ly, ry) + dh

        barx = [lx, lx, rx, rx]
        bary = [y, y+barh, y+barh, y]
        mid = ((lx+rx)/2, y+barh)

        ax.plot(np.array(bary)+delta, barx, alpha=0.8, clip_on=False)   
        
    '''
    Deprecated 14 March 2022. 
    def annotate_bars(self, ax, bottom_idx, top_idx, x=0, **kwargs):
        """
        Adds a square bracket that contains two points. Used to
        connect predictors in the predictor ranking plot
        for highly correlated pairs.
        """
        color = kwargs.get("color", "xkcd:slate gray")
        ax.annotate(
            "",
            xy=(x, bottom_idx),
            xytext=(x, top_idx),
            arrowprops=dict(
                arrowstyle="<->,head_length=0.05,head_width=0.05",
                ec=color,
                connectionstyle="bar,fraction=0.2",
                shrinkA=0.5,
                shrinkB=0.5,
                linewidth=0.5,
            ),
        )
    '''

    def get_custom_colormap(self, vals, **kwargs):
        """Get a custom colormap"""
        cmap = kwargs.get("cmap", matplotlib.cm.PuOr)
        bounds = np.linspace(np.nanpercentile(vals, 0), np.nanpercentile(vals, 100), 10)

        norm = matplotlib.colors.BoundaryNorm(
            bounds,
            cmap.N,
        )
        mappable = matplotlib.cm.ScalarMappable(
            norm=norm,
            cmap=cmap,
        )

        return mappable, bounds

    def add_ice_colorbar(self, fig, ax, mappable, cb_label, cdata, fontsize, **kwargs):
        """Add a colorbar to the right of a panel to
        accompany ICE color-coded plots"""
        cb = plt.colorbar(mappable, ax=ax, pad=0.2)
        cb.set_label(cb_label, size=fontsize)
        cb.ax.tick_params(labelsize=fontsize)
        cb.set_alpha(1)
        cb.outline.set_visible(False)
        bbox = cb.ax.get_window_extent().transformed(fig.dpi_scale_trans.inverted())
        cb.ax.set_aspect((bbox.height - 0.7) * 20)
        self._to_sci_notation(ax=None, colorbar=cb, ydata=cdata)

    def add_colorbar(
        self,
        fig,
        plot_obj,
        colorbar_label,
        ticks=MaxNLocator(5),
        ax=None,
        cax=None,
        **kwargs,
    ):
        """Adds a colorbar to the right of a panel"""
        # Add a colobar
        orientation = kwargs.get("orientation", "vertical")
        pad = kwargs.get("pad", 0.1)
        shrink = kwargs.get("shrink", 1.1)
        extend = kwargs.get("extend", "neither")

        if cax:
            cbar = plt.colorbar(
                plot_obj,
                cax=cax,
                pad=pad,
                ticks=ticks,
                shrink=shrink,
                orientation=orientation,
                extend=extend,
            )
        else:
            cbar = plt.colorbar(
                plot_obj,
                ax=ax,
                pad=pad,
                ticks=ticks,
                shrink=shrink,
                orientation=orientation,
                extend=extend,
            )
        cbar.ax.tick_params(labelsize=self.FONT_SIZES["tiny"])
        cbar.set_label(colorbar_label, size=self.FONT_SIZES["small"])
        cbar.outline.set_visible(False)
        # bbox = cbar.ax.get_window_extent().transformed(fig.dpi_scale_trans.inverted())
        # cbar.ax.set_aspect((bbox.height - 0.7) * 20)

    def save_figure(self, fname, fig=None, bbox_inches="tight", dpi=300, aformat="png"):
        """Saves the current figure"""
        plt.savefig(fname=fname, bbox_inches=bbox_inches, dpi=dpi, format=aformat)



"""While the various variable importance methods can, in general, for many 
different situations, such as evaluating the model-agnostic presence of 
information withing a dataset, the most typical application of the method is to
determine the importance of variables as evaluated by a particular model. The 
tools provide here are useful to assist in the training and evaluation of 
sklearn models. This is done by wrapping the training and evaluation of the 
model into a single function which is then used as the ``scoring_fn`` of a 
generalized variable importance method.

All of the variable importance methods with a ``sklearn_`` prefix use these 
tools to determine 1) whether to retrain a model at each step (as is necessary 
for Sequential Selection, but not for Permutation Importance) and 2) how to 
evaluate the resulting predictions of a model.

Here, the powerhouse is the ``model_scorer`` object, which handles all of the 
typical use-cases for any model by separately applying a training, prediction,
and evaluation function. Supplied with proper functions for each of these, the
``model_scorer`` object could also be implemented to score other types of 
models, such as Keras models."""

import numpy as np
from sklearn.base import clone

from .utils import get_data_subset, bootstrap_generator
from joblib import Parallel, delayed


__all__ = [
    "model_scorer",
    "score_untrained_sklearn_model",
    "score_untrained_sklearn_model_with_probabilities",
    "score_trained_sklearn_model",
    "score_trained_sklearn_model_with_probabilities",
    "train_model",
    "get_model",
    "predict_model",
    "predict_proba_model",
]


def train_model(model, X_train, y_train):
    """Trains a scikit-learn model and returns the trained model"""
    if X_train.shape[1] == 0:
        # No data to train over, so don't bother
        return None
    cloned_model = clone(model)
    return cloned_model.fit(X_train, y_train)


def get_model(model, X_train, y_train):
    """Just return the trained model"""
    return model


def predict_model(model, X_score):
    """Uses a trained scikit-learn model to predict over the scoring data"""
    return model.predict(X_score)


def predict_proba_model(model, X_score):
    """Uses a trained scikit-learn model to predict class probabilities for the
    scoring data"""
    return model.predict_proba(X_score)[:, 1]


def forward_permutations(X, inds, var_idx):
    return np.array(
                    [
                        X[:, i] if i == var_idx else X[inds, i]
                        for i in range(X.shape[1])
                    ]
                ).T

class model_scorer(object):
    """General purpose scoring method which takes a particular model, trains the
    model over the given training data, uses the trained model to predict on the
    given scoring data, and then evaluates those predictions using some
    evaluation function. Additionally provides the tools for bootstrapping the
    scores and providing a distribution of scores to be used for statistics.
    
    NOTE: Since these method is used internally, the scoring inputs into 
    this method for different rounds of multipass permutation importance 
    are already permuted for the top most features. Thus, in any current 
    iteration, we need only permute a single column at a time. 
    """
    def __init__(
        self,
        model,
        training_fn,
        prediction_fn,
        evaluation_fn,
        nimportant_vars=1,
        default_score=0.0,
        n_permute=1,
        subsample=1,
        direction='backward', 
        **kwargs
    ):
        """Initializes the scoring object by storing the training, predicting,
        and evaluation functions

        :param model: a scikit-learn model
        :param training_fn: a function for training a scikit-learn model. Must
            be of the form ``(model, X_train, y_train) ->
            trained_model | None``. If the function returns ``None``, then it is
            assumed that the model training failed.
            Probably :func:`PermutationImportance.sklearn_api.train_model` or
            :func:`PermutationImportance.sklearn_api.get_model`
        :param predicting_fn: a function for predicting on scoring data using a
            scikit-learn model. Must be of the form ``(model, X_score) ->
            predictions``. Predictions may be either deterministic or
            probabilistic, depending on what the evaluation_fn accepts.
            Probably :func:`PermutationImportance.sklearn_api.predict_model` or
            :func:`PermutationImportance.sklearn_api.predict_proba_model`
        :param evaluation_fn: a function which takes the deterministic or
            probabilistic model predictions and scores them against the true
            values. Must be of the form ``(truths, predictions) -> some_value``
            Probably one of the metrics in
            :mod:`PermutationImportance.metrics` or
            `sklearn.metrics <https://scikit-learn.org/stable/modules/classes.html#module-sklearn.metrics>`_
        :param default_score: value to return if the model cannot be trained
        :param nbootstrap: number of times to perform scoring on each variable.
            Results over different bootstrap iterations are averaged. Defaults
            to None, which will not perform bootstrapping
        :param subsample: number of elements to sample (with replacement) per
            bootstrap round. If between 0 and 1, treated as a fraction of the number
            of total number of events (e.g. 0.5 means half the number of events).
            If not specified, subsampling will not be used and the entire data will
            be used (without replacement)
        """

        self.model = model
        self.training_fn = training_fn
        self.prediction_fn = prediction_fn
        self.evaluation_fn = evaluation_fn
        self.default_score = default_score
        self.n_permute = n_permute
        self.subsample = subsample
        self.direction = direction
        self.kwargs = kwargs
        self.random_seed = kwargs.get("random_seed", 42)

    def _scorer(self, X, y):
        predictions = self.prediction_fn(self.model, X)
        return self.evaluation_fn(y,predictions)
    
    def get_subsample_size(self, full_size):
        return (
            int(full_size * self.subsample)
            if self.subsample <= 1
            else self.subsample
        )
    
    def _train(self):
        # Try to train model
        trained_model = self.training_fn(self.model, X_train, y_train)
        # If we didn't succeed in training (probably because there weren't any
        # training predictors), return the default_score
        if trained_model is None:
            if self.n_permute == 1:
                return [self.default_score]
            else:
                return np.full((self.n_permute,), self.default_score)
    
    
    def get_permuted_data(self, idx, var_idx):
        """ Get permuted data """
        X_score_sub = self.X_score
        y_score_sub = self.y_score
        X_train_sub = self.X_train
        
        inds = self.shuffled_indices[idx]
        
        if len(self.rows[0]) != self.y_score.shape[0]: 
            X_score_sub = get_data_subset(self.X_score, self.rows[idx])
            y_score_sub = get_data_subset(self.y_score, self.rows[idx])
        
            if self.direction == 'forward':
                X_train_sub = get_data_subset(self.X_train, self.rows[idx])
                
            #inds = inds[idx]
        
        if var_idx is None:
            return X_score_sub, y_score_sub
        
        # For the backward, X_score is mostly unpermuted expect for 
        # the top features. For the forward, X_score is all permuted 
        # expect for the top features. 
        X_perm = X_score_sub.copy()
        
        if self.direction == 'backward':
            X_perm[:,var_idx] = X_score_sub[inds, var_idx]
        else:
            X_perm[:,var_idx] = X_train_sub[:, var_idx]
            
        return X_perm, y_score_sub
        
    def __call__(self, training_data, scoring_data, var_idx):
        """Uses the training, predicting, and evaluation functions to score the
        model given the training and scoring data

        :param training_data: (training_input, training_output)
        :param scoring_data: (scoring_input, scoring_output)
        :returns: either a single value or an array of values
        :param var_idx : integer
            The column index of the variable being permuted. When computing the original
            score, set var_idx==None.
        """
        (self.X_train, self.y_train) = training_data
        (self.X_score, self.y_score) = scoring_data
        
        permuted_set = [self.get_permuted_data(idx, var_idx) for idx in range(self.n_permute)] 
        scores = np.array([self._scorer(*arg) for arg in permuted_set]) 
    
        return np.array(scores) 
    

def score_untrained_sklearn_model(
    model, evaluation_fn, nbootstrap=None, subsample=1, **kwargs
):
    """A convenience method which uses the default training and the
    deterministic prediction methods for scikit-learn to evaluate a model

    :param model: a scikit-learn model
    :param evaluation_fn: a function which takes the deterministic or
        probabilistic model predictions and scores them against the true
        values. Must be of the form ``(truths, predictions) -> some_value``
        Probably one of the metrics in
        :mod:`PermutationImportance.metrics` or
        `sklearn.metrics <https://scikit-learn.org/stable/modules/classes.html#module-sklearn.metrics>`_
    :param nbootstrap: number of times to perform scoring on each variable.
        Results over different bootstrap iterations are averaged. Defaults to 1
    :param subsample: number of elements to sample (with replacement) per
        bootstrap round. If between 0 and 1, treated as a fraction of the number
        of total number of events (e.g. 0.5 means half the number of events).
        If not specified, subsampling will not be used and the entire data will
        be used (without replacement)
    :param kwargs: all other kwargs passed on to the evaluation_fn
    :returns: a callable which accepts ``(training_data, scoring_data)`` and
        returns some value (probably a float or an array of floats)
    """
    return model_scorer(
        model,
        training_fn=train_model,
        prediction_fn=predict_model,
        evaluation_fn=evaluation_fn,
        nbootstrap=nbootstrap,
        subsample=subsample,
        **kwargs
    )


def score_untrained_sklearn_model_with_probabilities(
    model, evaluation_fn, nbootstrap=None, subsample=1, **kwargs
):
    """A convenience method which uses the default training and the
    probabilistic prediction methods for scikit-learn to evaluate a model

    :param model: a scikit-learn model
    :param evaluation_fn: a function which takes the deterministic or
        probabilistic model predictions and scores them against the true
        values. Must be of the form ``(truths, predictions) -> some_value``
        Probably one of the metrics in
        :mod:`PermutationImportance.metrics` or
        `sklearn.metrics <https://scikit-learn.org/stable/modules/classes.html#module-sklearn.metrics>`_
    :param nbootstrap: number of times to perform scoring on each variable.
        Results over different bootstrap iterations are averaged. Defaults to 1
    :param subsample: number of elements to sample (with replacement) per
        bootstrap round. If between 0 and 1, treated as a fraction of the number
        of total number of events (e.g. 0.5 means half the number of events).
        If not specified, subsampling will not be used and the entire data will
        be used (without replacement)
    :param kwargs: all other kwargs passed on to the evaluation_fn
    :returns: a callable which accepts ``(training_data, scoring_data)`` and
        returns some value (probably a float or an array of floats)
    """
    return model_scorer(
        model,
        training_fn=train_model,
        prediction_fn=predict_proba_model,
        evaluation_fn=evaluation_fn,
        nbootstrap=nbootstrap,
        subsample=subsample,
        **kwargs
    )


def score_trained_sklearn_model(
    model, evaluation_fn, n_permute=1, subsample=1, direction='backward', **kwargs
):
    """A convenience method which does not retrain a scikit-learn model and uses
    deterministic prediction methods to evaluate the model

    :param model: a scikit-learn model
    :param evaluation_fn: a function which takes the deterministic or
        probabilistic model predictions and scores them against the true
        values. Must be of the form ``(truths, predictions) -> some_value``
        Probably one of the metrics in
        :mod:`PermutationImportance.metrics` or
        `sklearn.metrics <https://scikit-learn.org/stable/modules/classes.html#module-sklearn.metrics>`_
    :param nbootstrap: number of times to perform scoring on each variable.
        Results over different bootstrap iterations are averaged. Defaults to 1
    :param subsample: number of elements to sample (with replacement) per
        bootstrap round. If between 0 and 1, treated as a fraction of the number
        of total number of events (e.g. 0.5 means half the number of events).
        If not specified, subsampling will not be used and the entire data will
        be used (without replacement)
    :param kwargs: all other kwargs passed on to the evaluation_fn
    :returns: a callable which accepts ``(training_data, scoring_data)`` and
        returns some value (probably a float or an array of floats)
    """
    return model_scorer(
        model,
        training_fn=get_model,
        prediction_fn=predict_model,
        evaluation_fn=evaluation_fn,
        n_permute=n_permute,
        subsample=subsample,
        direction=direction, 
        **kwargs
    )


def score_trained_sklearn_model_with_probabilities(
    model, evaluation_fn, n_permute=1, subsample=1, direction='backward', **kwargs
):
    """A convenience method which does not retrain a scikit-learn model and uses
    probabilistic prediction methods to evaluate the model

    :param model: a scikit-learn model
    :param evaluation_fn: a function which takes the deterministic or
        probabilistic model predictions and scores them against the true
        values. Must be of the form ``(truths, predictions) -> some_value``
        Probably one of the metrics in
        :mod:`PermutationImportance.metrics` or
        `sklearn.metrics <https://scikit-learn.org/stable/modules/classes.html#module-sklearn.metrics>`_
    :param nbootstrap: number of times to perform scoring on each variable.
        Results over different bootstrap iterations are averaged. Defaults to 1
    :param subsample: number of elements to sample (with replacement) per
        bootstrap round. If between 0 and 1, treated as a fraction of the number
        of total number of events (e.g. 0.5 means half the number of events).
        If not specified, subsampling will not be used and the entire data will
        be used (without replacement)
    :param kwargs: all other kwargs passed on to the evaluation_fn
    :returns: a callable which accepts ``(training_data, scoring_data)`` and
        returns some value (probably a float or an array of floats)
    """
    return model_scorer(
        model,
        training_fn=get_model,
        prediction_fn=predict_proba_model,
        evaluation_fn=evaluation_fn,
        n_permute=n_permute,
        subsample=subsample,
        direction=direction,
        **kwargs
    )



from functools import partial
from sklearn.metrics._base import _average_binary_score
from sklearn.utils.multiclass import type_of_target
from sklearn.metrics import (
    brier_score_loss,
    average_precision_score,
    precision_recall_curve,
)
import numpy as np


def brier_skill_score(y_values, forecast_probabilities):
    """Computes the brier skill score"""
    climo = np.mean((y_values - np.mean(y_values)) ** 2)
    return 1.0 - brier_score_loss(y_values, forecast_probabilities) / climo


def modified_precision(precision, known_skew, new_skew):
    """
    Modify the success ratio according to equation (3) from
    Lampert and Gancarski (2014).
    """
    precision[precision < 1e-5] = 1e-5

    term1 = new_skew / (1.0 - new_skew)
    term2 = (1 / precision) - 1.0

    denom = known_skew + ((1 - known_skew) * term1 * term2)

    return known_skew / denom


def calc_sr_min(skew):
    pod = np.linspace(0, 1, 100)
    sr_min = (skew * pod) / (1 - skew + (skew * pod))
    return sr_min


def _binary_uninterpolated_average_precision(
    y_true, y_score, known_skew, new_skew, pos_label=1, sample_weight=None
):
    precision, recall, _ = precision_recall_curve(
        y_true, y_score, pos_label=pos_label, sample_weight=sample_weight
    )
    # Return the step function integral
    # The following works because the last entry of precision is
    # guaranteed to be 1, as returned by precision_recall_curve
    if known_skew is not None:
        precision = modified_precision(precision, known_skew, new_skew)
    return -np.sum(np.diff(recall) * np.array(precision)[:-1])


def min_aupdc(
    y_true, pos_label, average, sample_weight=None, known_skew=None, new_skew=None
):
    """
    Compute the minimum possible area under the performance
    diagram curve. Essentially, a vote of NO for all predictions.
    """
    min_score = np.zeros((len(y_true)))
    average_precision = partial(
        _binary_uninterpolated_average_precision,
        known_skew=known_skew,
        new_skew=new_skew,
        pos_label=pos_label,
    )
    ap_min = _average_binary_score(
        average_precision, y_true, min_score, average, sample_weight=sample_weight
    )

    return ap_min


def norm_aupdc(
    y_true,
    y_score,
    known_skew=None,
    *,
    average="macro",
    pos_label=1,
    sample_weight=None,
    min_method="random",
):
    """
    Compute the normalized modified average precision. Normalization removes
    the no-skill region either based on skew or random classifier performance.
    Modification alters success ratio to be consistent with a known skew.

    Parameters:
    -------------------
        y_true, array of (n_samples,)
            Binary, truth labels (0,1)
        y_score, array of (n_samples,)
            Model predictions (either determinstic or probabilistic)
        known_skew, float between 0 and 1
            Known or reference skew (# of 1 / n_samples) for
            computing the modified success ratio.
        min_method, 'skew' or 'random'
            If 'skew', then the normalization is based on the minimum AUPDC
            formula presented in Boyd et al. (2012).

            If 'random', then the normalization is based on the
            minimum AUPDC for a random classifier, which is equal
            to the known skew.


    Boyd, 2012: Unachievable Region in Precision-Recall Space and Its Effect on Empirical Evaluation, ArXiv
    """
    new_skew = np.mean(y_true)
    if known_skew is None:
        known_skew = new_skew

    y_type = type_of_target(y_true)
    if y_type == "multilabel-indicator" and pos_label != 1:
        raise ValueError(
            "Parameter pos_label is fixed to 1 for "
            "multilabel-indicator y_true. Do not set "
            "pos_label or set pos_label to 1."
        )
    elif y_type == "binary":
        # Convert to Python primitive type to avoid NumPy type / Python str
        # comparison. See https://github.com/numpy/numpy/issues/6784
        present_labels = np.unique(y_true).tolist()
        if len(present_labels) == 2 and pos_label not in present_labels:
            raise ValueError(
                f"pos_label={pos_label} is not a valid label. It should be "
                f"one of {present_labels}"
            )
    average_precision = partial(
        _binary_uninterpolated_average_precision,
        known_skew=known_skew,
        new_skew=new_skew,
        pos_label=pos_label,
    )

    ap = _average_binary_score(
        average_precision, y_true, y_score, average, sample_weight=sample_weight
    )

    if min_method == "random":
        ap_min = known_skew
    elif min_method == "skew":
        ap_min = min_aupdc(
            y_true,
            pos_label,
            average,
            sample_weight=sample_weight,
            known_skew=known_skew,
            new_skew=new_skew,
        )

    naupdc = (ap - ap_min) / (1.0 - ap_min)

    return naupdc



"""Permutation Importance determines which variables are important by comparing
performance on a dataset where some of the variables are permuted in their 
individual columns to performance on the dataset without any permutation. The
permutation of an individual variable in this manner has the effect of breaking
any relationship between the input variable and the target. The variable which,
when permuted, results in the worst performance is typically taken as the most
important variable.

Typically, when using a performance metric or skill score with Permutation 
Importance, the ``scoring_strategy`` should be to minimize the performance. On 
the other hand, when using an error or loss function, the ``scoring_strategy`` 
should be to maximize the error or loss function."""

import numpy as np

from .abstract_runner import abstract_variable_importance
from .selection_strategies import (
    PermutationImportanceSelectionStrategy,
    ConditionalPermutationImportanceSelectionStrategy,
    ForwardPermutationImportanceSelectionStrategy,
)
from .sklearn_api import (
    score_trained_sklearn_model,
    score_trained_sklearn_model_with_probabilities,
)

__all__ = ["permutation_importance", "sklearn_permutation_importance"]

# fake_data = np.zeros((3,3))
# variable_names = ['' for i in range(len(fake_data))]
# from sklearn.metrics import roc_auc_score


def permutation_importance(
    scoring_data,
    scoring_fn,
    scoring_strategy,
    variable_names=None,
    nimportant_vars=None,
    njobs=1,
    direction="backward",
    verbose=False,
    random_seed=1,
    **kwargs,
):
    """Performs permutation importance over data given a particular
    set of functions for scoring and determining optimal variables

    :param scoring_data: a 2-tuple ``(inputs, outputs)`` for scoring in the
        ``scoring_fn``
    :param scoring_fn: a function to be used for scoring. Should be of the form
        ``(training_data, scoring_data) -> some_value``
    :param scoring_strategy: a function to be used for determining optimal
        variables. Should be of the form ``([some_value]) -> index``
    :param variable_names: an optional list for variable names. If not given,
        will use names of columns of data (if pandas dataframe) or column
        indices
    :param nimportant_vars: number of variables to compute multipass importance
        for. Defaults to all variables
    :param njobs: an integer for the number of threads to use. If negative, will
        use ``num_cpus + njobs``. Defaults to 1
    :param direction: 'forward' or 'backward': Whether the top feature is left permuted (backward)
        or all features are permuted and the top features are progressively left unpermuted (forward).
    :returns: :class:`PermutationImportance.result.ImportanceResult` object
        which contains the results for each run
    """
    if direction == "conditional":
        selection_strategy = ConditionalPermutationImportanceSelectionStrategy
    elif direction == "backward":
        selection_strategy = PermutationImportanceSelectionStrategy
    elif direction == "forward":
        selection_strategy = ForwardPermutationImportanceSelectionStrategy

    else:
        raise ValueError(f'method must be "conditional", "forward", or "backward"!')

    # We don't need the training data, so pass empty arrays to the abstract runner
    if scoring_data is None:
        raise ValueError("Must declare scoring data!")
    else:
        return abstract_variable_importance(
            training_data=(np.array([]), np.array([])),
            scoring_data=scoring_data,
            scoring_fn=scoring_fn,
            scoring_strategy=scoring_strategy,
            selection_strategy=selection_strategy,
            variable_names=variable_names,
            nimportant_vars=nimportant_vars,
            njobs=njobs,
            verbose=verbose,
            random_seed=random_seed,
            direction=direction,
        )

def sklearn_permutation_importance(
    model,
    scoring_data,
    evaluation_fn,
    scoring_strategy,
    variable_names=None,
    nimportant_vars=None,
    direction="backward",
    njobs=1,
    n_permute=1,
    subsample=1,
    verbose=False,
    random_seed=1,
    **scorer_kwargs,
):

    """Performs permutation importance for a particular model,
    ``scoring_data``, ``evaluation_fn``, and strategy for determining optimal
    variables

    :param model: a trained sklearn model
    :param scoring_data: a 2-tuple ``(inputs, outputs)`` for scoring in the
        ``scoring_fn``
    :param evaluation_fn: a function which takes the deterministic or
        probabilistic model predictions and scores them against the true
        values. Must be of the form ``(truths, predictions) -> some_value``
        Probably one of the metrics in
        :mod:`PermutationImportance.metrics` or
        `sklearn.metrics <https://scikit-learn.org/stable/modules/classes.html#module-sklearn.metrics>`_
    :param scoring_strategy: a function to be used for determining optimal
    variables. Should be of the form ``([some_value]) -> index``
    :param variable_names: an optional list for variable names. If not given,
        will use names of columns of data (if pandas dataframe) or column
        indices
    :param nimportant_vars: number of variables to compute multipass importance
        for. Defaults to all variables
    :param njobs: an integer for the number of threads to use. If negative, will
        use ``num_cpus + njobs``. Defaults to 1
    :param n_permute: number of times to perform permutation on each variable.
        Results over different permutations iterations are averaged. Defaults to 1
    :param subsample: number of elements to sample (with replacement) per
        bootstrap round. If between 0 and 1, treated as a fraction of the number
        of total number of events (e.g. 0.5 means half the number of events).
        If not specified, subsampling will not be used and the entire data will
        be used (without replacement)
    :param kwargs: all other kwargs will be passed on to the ``evaluation_fn``
    :returns: :class:`PermutationImportance.result.ImportanceResult` object
        which contains the results for each run
    """
    # Check if the data is probabilistic
    # if len(scoring_data[1].shape) > 1 and scoring_data[1].shape[1] > 1:
    if len(np.unique(scoring_data[1])) == 2:
        scoring_fn = score_trained_sklearn_model_with_probabilities(
            model,
            evaluation_fn,
            n_permute=n_permute,
            subsample=subsample,
            random_seed=random_seed,
            direction=direction,
            scorer_kwargs={},
        )
    else:
        scoring_fn = score_trained_sklearn_model(
            model,
            evaluation_fn,
            n_permute=n_permute,
            subsample=subsample,
            random_seed=random_seed,
            direction=direction,
            scorer_kwargs={},
        )

    return permutation_importance(
        scoring_data=scoring_data,
        scoring_fn=scoring_fn,
        scoring_strategy=scoring_strategy,
        variable_names=variable_names,
        nimportant_vars=nimportant_vars,
        njobs=njobs,
        verbose=verbose,
        direction=direction,
        random_seed=random_seed,
        scorer_kwargs={},
    )



import numpy as np
import collections

from ..common.importance_utils import find_correlated_pairs_among_top_features
from ..common.utils import is_list, is_correlated
from .base_plotting import PlotStructure
import random


class PlotImportance(PlotStructure):
    """
    PlotImportance handles plotting feature ranking plotting. The class
    is designed to be generic enough to handle all possible ranking methods
    computed within PyMint.
    """

    SINGLE_VAR_METHODS = [
        "multipass",
        "singlepass",
        "ale_variance",
        "coefs",
        "shap_sum",
        "gini",
        "combined",
        "sage",
        "grouped",
        "grouped_only",
    ]

    DISPLAY_NAMES_DICT = {
        "multipass": "Multi-Pass",
        "singlepass": "Single-Pass",
        "perm_based": "Permutation-based Interactions",
        "ale_variance": "ALE-Based Importance",
        "ale_variance_interactions": "ALE-Based Interactions",
        "coefs": "Coef.",
        "shap_sum": "SHAP",
        "hstat": "H-Stat",
        "gini": "Gini",
        "combined": "Method-Average Ranking",
        "sage": "SAGE Importance Scores",
        "grouped": "Grouped Importance",
        "grouped_only": "Grouped Only Importance",
    }

    def __init__(self, BASE_FONT_SIZE=12):
        super().__init__(BASE_FONT_SIZE=BASE_FONT_SIZE)

    def is_bootstrapped(self, scores):
        """Check if the permutation importance results are bootstrapped"""
        return np.ndim(scores) > 1

    def _get_axes(self, n_panels, **kwargs):
        """
        Determine how many axes are required.
        """
        if n_panels == 1:
            kwargs["figsize"] = kwargs.get("figsize", (3, 2.5))
        elif n_panels == 2 or n_panels == 3:
            kwargs["figsize"] = kwargs.get("figsize", (6, 2.5))
        else:
            figsize = kwargs.get("figsize", (8, 5))

        # create subplots, one for each feature
        fig, axes = self.create_subplots(n_panels=n_panels, **kwargs)

        return fig, axes

    def _check_for_estimators(self, data, estimator_names):
        """Check that each estimator is in data"""
        for ds in data:
            if not (
                collections.Counter(ds.attrs["estimators used"])
                == collections.Counter(estimator_names)
            ):
                raise AttributeError(
                    """
                 The estimator names given do not match the estimators used to create
                 given data 
                                     """
                )

    def plot_variable_importance(
        self,
        data,
        panels,
        display_feature_names={},
        feature_colors=None,
        num_vars_to_plot=10,
        estimator_output="raw",
        plot_correlated_features=False,
        **kwargs,
    ):

        """Plots any variable importance method for a particular estimator

        Parameters
        -----------------
            data : xarray.Dataset or list of xarray.Dataset
                Permutation importance dataset for one or more metrics
            panels: list of 2-tuples of estimator names and rank method
                E.g., panels = [('singlepass', 'Random Forest',
                                ('multipass', 'Random Forest') ]
                will plot the singlepass and multipass results for the
                random forest model.
                Possible methods include 'multipass', 'singlepass',
                'perm_based', 'ale_variance', or 'ale_variance_interactions'
            display_feature_names : dict
                A dict mapping feature names to readable, "pretty" feature names
            feature_colors : dict
                A dict mapping features to various colors. Helpful for color coding groups of features
            num_vars_to_plot : int
                Number of top variables to plot (defalut is None and will use number of multipass results)
                
            kwargs: 
                - xlabels 
                - ylabel 
                - xticks
                - p_values 
                - colinear_features 
                - rho_threshold 
                
        """
        xlabels = kwargs.get("xlabels", None)
        ylabels = kwargs.get("ylabels", None)
        xticks = kwargs.get("xticks", None)
        title = kwargs.get("title", "")
        p_values = kwargs.get("p_values", None)
        colinear_features = kwargs.get("colinear_features", None)
        rho_threshold = kwargs.get("rho_threshold", 0.8)

        only_one_method = all([m[0] == panels[0][0] for m in panels])
        only_one_estimator = all([m[1] == panels[0][1] for m in panels])

        if not only_one_method:
            kwargs["hspace"] = kwargs.get("hspace", 0.6)

        if plot_correlated_features:
            X = kwargs.get("X", None)
            if X is None or X.empty:
                raise ValueError(
                    "Must provide X to InterpretToolkit to compute the correlations!"
                )
            corr_matrix = X.corr().abs()

        data = [data] if not is_list(data) else data
        n_panels = len(panels)

        fig, axes = self._get_axes(n_panels, **kwargs)
        ax_iterator = self.axes_to_iterator(n_panels, axes)

        for i, (panel, ax) in enumerate(zip(panels, ax_iterator)):

            # Set the facecolor.
            #ax.set_facecolor(kwargs.get("facecolor", (0.95, 0.95, 0.95)))

            method, estimator_name = panel
            results = data[i]
            if xlabels is not None:
                ax.set_xlabel(xlabels[i], fontsize=self.FONT_SIZES["small"])
            else:
                if not only_one_method:
                    ax.set_xlabel(
                        self.DISPLAY_NAMES_DICT.get(method, method),
                        fontsize=self.FONT_SIZES["small"],
                    )
            if not only_one_estimator:
                ax.set_title(estimator_name)

            sorted_var_names = list(
                results[f"{method}_rankings__{estimator_name}"].values
            )

            if num_vars_to_plot is None:
                num_vars_to_plot == len(sorted_var_names)

            sorted_var_names = sorted_var_names[
                : min(num_vars_to_plot, len(sorted_var_names))
            ]

            sorted_var_names = sorted_var_names[::-1]
            scores = results[f"{method}_scores__{estimator_name}"].values

            scores = scores[: min(num_vars_to_plot, len(sorted_var_names))]

            # Reverse the order.
            scores = scores[::-1]

            # Set very small values to zero.
            scores = np.where(np.absolute(np.round(scores, 17)) < 1e-15, 0, scores)

            # Get the colors for the plot
            colors_to_plot = [
                self.variable_to_color(var, feature_colors) for var in sorted_var_names
            ]
            
            # Get the predictor names
            variable_names_to_plot = [
                f" {var}"
                for var in self.convert_vars_to_readable(
                    sorted_var_names,
                    display_feature_names,
                )
            ]

            if method == "combined":
                scores_to_plot = np.nanpercentile(scores, 50, axis=1)
                # Compute the confidence intervals (ci)
                ci = np.abs(
                    np.nanpercentile(scores, 50, axis=1)
                    - np.nanpercentile(scores, [25, 75], axis=1)
                )
            else:
                scores_to_plot = np.nanmean(scores, axis=1)
                ci = np.abs(
                    np.nanpercentile(scores, 50, axis=1)
                    - np.nanpercentile(scores, [2.5, 97.5], axis=1)
                )

            # Despine
            self.despine_plt(ax)

            elinewidth = 0.9 if n_panels <= 3 else 0.5

            ax.barh(
                np.arange(len(scores_to_plot)),
                scores_to_plot,
                linewidth=1.75,
                edgecolor="white",
                alpha=0.5,
                color=colors_to_plot,
                xerr=ci,
                capsize=3.0,
                ecolor="k",
                error_kw=dict(
                    alpha=0.2,
                    elinewidth=elinewidth,
                ),
                zorder=2,
            )

            
            if plot_correlated_features:
                self._add_correlated_brackets(
                    ax, np.arange(len(scores_to_plot)), 
                    scores_to_plot,
                    corr_matrix, sorted_var_names, rho_threshold
                )

            if num_vars_to_plot >= 20:
                size = kwargs.get("fontsize", self.FONT_SIZES["teensie"] - 3)
            elif num_vars_to_plot > 10:
                size = kwargs.get("fontsize", self.FONT_SIZES["teensie"] - 2)
            else:
                size = kwargs.get("fontsize", self.FONT_SIZES["teensie"] - 1)


            # Put the variable names _into_ the plot
            if method not in self.SINGLE_VAR_METHODS and plot_correlated_features:
                results_dict = is_correlated(
                    corr_matrix, sorted_var_names, rho_threshold=rho_threshold
                )

            if colinear_features is None:
                fontweight = ["light"] * len(variable_names_to_plot)
                colors = ["k"] * len(variable_names_to_plot)
            else:
                # Bold text if the VIF > threshold (indicates a multicolinear predictor)
                fontweight =  [
                    "bold" if v in colinear_features else "light" for v in sorted_var_names
                ]  
                
                # Bold text if value is insignificant.
                colors =  ["xkcd:medium blue" if v in colinear_features else "k" for v in sorted_var_names]
                

            ax.set_yticks(range(len(variable_names_to_plot)))
            ax.set_yticklabels(variable_names_to_plot)
            labels = ax.get_yticklabels()
            
            # Bold var names 
            ##[label.set_fontweight(opt) for opt, label in zip(fontweight, labels)]
            
            [label.set_color(c) for c, label in zip(colors, labels)]
            
            ax.tick_params(axis="both", which="both", length=0)

            if xticks is not None:
                ax.set_xticks(xticks)
            else:
                self.set_n_ticks(ax, option="x")
                
        xlabel = (
            self.DISPLAY_NAMES_DICT.get(method, method)
            if (only_one_method and xlabels is None)
            else ""
        )
        
        major_ax = self.set_major_axis_labels(
            fig,
            xlabel=xlabel,
            ylabel_left="",
            ylabel_right="",
            title=title,
            fontsize=self.FONT_SIZES["small"],
            **kwargs,
        )
        
        if ylabels is not None:
            self.set_row_labels(
                labels=ylabels, axes=axes, pos=-1, pad=1.15, rotation=270, **kwargs
            )

        self.add_alphabet_label(
            n_panels, axes, pos=kwargs.get("alphabet_pos", (0.9, 0.09)), 
            alphabet_fontsize = kwargs.get("alphabet_fontsize", 10)
        )

        # Necessary to make sure that the tick labels for the feature names
        # do overlap another ax. 
        fig.tight_layout()
        
        return fig, axes

    def _add_correlated_brackets(self, ax, y, width, corr_matrix, top_features, rho_threshold):
        """
        Add bracket connecting features above a given correlation threshold.
        
        Parameters
        ------------------
        ax : matplotlib.ax.Axes object 
        y : 
        width : 
        corr_matrix: 
        top_features:
        rho_threshold:
        """
        get_colors = lambda n: list(
            map(lambda i: "#" + "%06x" % random.randint(0, 0xFFFFFF), range(n))
        )

        _, pair_indices = find_correlated_pairs_among_top_features(
            corr_matrix,
            top_features,
            rho_threshold=rho_threshold,
        )
        
        colors = get_colors(len(pair_indices))

        top_indices, bottom_indices = [], []
        for p, color in zip(pair_indices, colors):
            delta=0
            if p[0] > p[1]:
                bottom_idx = p[1]
                top_idx = p[0]
            else:
                bottom_idx = p[0]
                top_idx = p[1]
            
            # If a feature has already shown up in a correlated pair,
            # then we want to shift the brackets slightly for ease of 
            # interpretation. 
            if bottom_idx in bottom_indices or bottom_idx in top_indices:
                delta += 0.1
            if top_idx in top_indices or top_idx in bottom_indices:
                delta += 0.1
                
            top_indices.append(top_idx)
            bottom_indices.append(bottom_idx)
            
            self.annotate_bars(ax, bottom_idx, top_idx, y=y, width=width, delta=delta)

    # You can fill this in by using a dictionary with {var_name: legible_name}
    def convert_vars_to_readable(self, variables_list, VARIABLE_NAMES_DICT):
        """Substitutes out variable names for human-readable ones
        :param variables_list: a list of variable names
        :returns: a copy of the list with human-readable names
        """
        human_readable_list = list()
        for var in variables_list:
            if var in VARIABLE_NAMES_DICT:
                human_readable_list.append(VARIABLE_NAMES_DICT[var])
            else:
                human_readable_list.append(var)
        return human_readable_list

    # This could easily be expanded with a dictionary
    def variable_to_color(self, var, VARIABLES_COLOR_DICT):
        """
        Returns the color for each variable.
        """
        if var == "No Permutations":
            return "xkcd:pastel red"
        else:
            if VARIABLES_COLOR_DICT is None:
                return "xkcd:powder blue"
            elif not isinstance(VARIABLES_COLOR_DICT, dict) and isinstance(
                VARIABLES_COLOR_DICT, str
            ):
                return VARIABLES_COLOR_DICT
            else:
                return VARIABLES_COLOR_DICT[var]



import os

current_dir = os.getcwd()
import pandas as pd
from os.path import join

path = os.path.dirname(os.path.realpath(__file__))


def load_data():
    """Loads road surface temperature dataset from Handler et al. (2020)"""
    cols_to_use = [
        "dllwave_flux",
        "dwpt2m",
        "fric_vel",
        "gflux",
        "high_cloud",
        "lat_hf",
        "low_cloud",
        "mid_cloud",
        "sat_irbt",
        "sens_hf",
        "sfcT_hrs_ab_frez",
        "sfcT_hrs_bl_frez",
        "sfc_rough",
        "sfc_temp",
        "swave_flux",
        "temp2m",
        "tmp2m_hrs_ab_frez",
        "tmp2m_hrs_bl_frez",
        "tot_cloud",
        "uplwav_flux",
        "vbd_flux",
        "vdd_flux",
        "wind10m",
        "date_marker",
        "urban",
        "rural",
        "d_ground",
        "d_rad_d",
        "d_rad_u",
        "hrrr_dT",
    ]

    # Import the training dataset
    data_filename = join(path, "data", "road_surface_data.csv")

    # Load the examples the models were trained on.
    TARGET_COLUMN = "cat_rt"
    data = pd.read_csv(data_filename)

    examples = data[cols_to_use]
    targets = data[TARGET_COLUMN].values

    return examples, targets



"""Various and sundry useful functions which are handy for manipulating data or
results of the variable importance"""

import numpy as np
import pandas as pd
import numbers

from .error_handling import InvalidDataException

__all__ = ["add_ranks_to_dict", "get_data_subset", "make_data_from_columns"]


def add_ranks_to_dict(result, variable_names, scoring_strategy):
    """Takes a list of (var, score) and converts to a dictionary of
    {var: (rank, score)}
    :param result: a dict of {var_index: score}
    :param variable_names: a list of variable names
    :param scoring_strategy: a function to be used for determining optimal
        variables. Should be of the form ([floats]) -> index
    """
    if len(result) == 0:
        return dict()

    result_dict = dict()
    rank = 0
    while len(result) > 1:
        var_idxs = list(result.keys())
        idxs = np.argsort(var_idxs)

        # Sort by indices to guarantee order
        variables = list(np.array(var_idxs)[idxs])
        scores = list(np.array(list(result.values()))[idxs])
        best_var = variables[scoring_strategy(scores)]
        score = result.pop(best_var)
        result_dict[variable_names[best_var]] = (rank, score)
        rank += 1
    var, score = list(result.items())[0]
    result_dict[variable_names[var]] = (rank, score)
    return result_dict


def get_data_subset(data, rows=None, columns=None):
    """Returns a subset of the data corresponding to the desired rows and
    columns
    :param data: either a pandas dataframe or a numpy array
    :param rows: a list of row indices
    :param columns: a list of column indices
    :returns: data_subset (same type as data)
    """
    if rows is None:
        rows = np.arange(data.shape[0])

    if isinstance(data, pd.DataFrame):
        if columns is None:
            return data.iloc[rows]
        else:
            return data.iloc[rows, columns]
    elif isinstance(data, np.ndarray):
        if columns is None:
            return data[rows]
        else:
            return data[np.ix_(rows, columns)]
    else:
        raise InvalidDataException(
            data, "Data must be a pandas dataframe or numpy array"
        )


def make_data_from_columns(columns_list, index=None):
    """Synthesizes a dataset out of a list of columns
    :param columns_list: a list of either pandas series or numpy arrays
    :returns: a pandas dataframe or a numpy array
    """
    if len(columns_list) == 0:
        raise InvalidDataException(
            columns_list, "Must have at least one column to synthesize dataset"
        )
    if isinstance(columns_list[0], pd.DataFrame) or isinstance(
        columns_list[0], pd.Series
    ):
        df = pd.concat([c.reset_index(drop=True) for c in columns_list], axis=1)
        if index is not None:
            return df.set_index(index)
        else:
            return df
    elif isinstance(columns_list[0], np.ndarray):
        return np.column_stack(columns_list)
    else:
        raise InvalidDataException(
            columns_list,
            "Columns_list must come from a pandas dataframe or numpy arrays",
        )


def conditional_permutations(data, n_bins, random_state):
    """
    Conditionally permute each feature in a dataset.

    Code appended to the PermutationImportance package by Montgomery Flora 2021.

    Args:
    -------------------
        data : pd.DataFrame or np.ndarray shape=(n_examples, n_features,)
        n_bins : interger
            number of bins to divide a feature into. Based on a
            percentile method to ensure that each bin receieves
            a similar number of examples
        random_state : np.random.RandomState instance
            Pseudo-random number generator to control the permutations of each
            feature.
            Pass an int to get reproducible results across function calls.

    Returns:
    -------------------
        permuted_data : a permuted version of data
    """
    permuted_data = data.copy()

    for i in range(np.shape(data)[1]):
        # Get the bin values of feature
        if isinstance(data, pd.DataFrame):
            feature_values = data.iloc[:, i]
        elif isinstance(data, np.ndarray):
            feature_values = data[:, i]
        else:
            raise InvalidDataException(
                data, "Data must be a pandas dataframe or numpy array"
            )

        bin_edges = np.unique(
            np.percentile(
                feature_values,
                np.linspace(0, 100, n_bins + 1),
                interpolation="lower",
            )
        )

        bin_indices = np.clip(
            np.digitize(feature_values, bin_edges, right=True) - 1, 0, None
        )

        shuffled_indices = bin_indices.copy()
        unique_bin_values = np.unique(bin_indices)

        # bin_indices is composed of bin indices for a corresponding value of feature_values
        for bin_idx in unique_bin_values:
            # idx is the actual index of indices where the bin index == i
            idx = np.where(bin_indices == bin_idx)[0]
            # Replace the bin indices with a permutation of the actual indices
            shuffled_indices[idx] = random_state.permutation(idx)

        if isinstance(data, pd.DataFrame):
            permuted_data.iloc[:, i] = data.iloc[shuffled_indices, i]
        else:
            permuted_data[:, i] = data[shuffled_indices, i]

    return permuted_data


def check_random_state(seed):
    """Turn seed into a np.random.RandomState instance
    Parameters. Function comes for sci-kit-learn.
    ----------
    seed : None, int or instance of RandomState
        If seed is None, return the RandomState singleton used by np.random.
        If seed is an int, return a new RandomState instance seeded with seed.
        If seed is already a RandomState instance, return it.
        Otherwise raise ValueError.
    """
    if seed is None or seed is np.random:
        return np.random.mtrand._rand
    if isinstance(seed, numbers.Integral):
        return np.random.RandomState(seed)
    if isinstance(seed, np.random.RandomState):
        return seed
    raise ValueError(
        "%r cannot be used to seed a numpy.random.RandomState" " instance" % seed
    )


def bootstrap_generator(n_bootstrap, seed=42):
    """
    Create a repeatable bootstrap generator.
    Will create the same set of random state generators given
    a number of bootstrap iterations.
    """
    base_random_state = np.random.RandomState(seed)
    random_num_set = base_random_state.choice(10000, size=n_bootstrap, replace=False)

    random_states = [np.random.RandomState(s) for s in random_num_set]

    return random_states



import matplotlib.pyplot as plt
import seaborn as sns
from matplotlib.colors import ListedColormap
from scipy.stats import gaussian_kde
from scipy.ndimage import gaussian_filter
import scipy
import itertools
import numpy as np
import matplotlib as mpl
from scipy.ndimage import gaussian_filter
from .base_plotting import PlotStructure

gray4 = (189 / 255.0, 189 / 255.0, 189 / 255.0)
gray5 = (150 / 255.0, 150 / 255.0, 150 / 255.0)
blue2 = (222 / 255.0, 235 / 255.0, 247 / 255.0)
blue5 = (107 / 255.0, 174 / 255.0, 214 / 255.0)
orange3 = (253 / 255.0, 208 / 255.0, 162 / 255.0)
orange5 = (253 / 255.0, 141 / 255.0, 60 / 255.0)
red5 = (251 / 255.0, 106 / 255.0, 74 / 255.0)
red6 = (239 / 255.0, 59 / 255.0, 44 / 255.0)
purple5 = (158 / 255.0, 154 / 255.0, 200 / 255.0)
purple6 = (128 / 255.0, 125 / 255.0, 186 / 255.0)
purple9 = (63 / 255.0, 0 / 255.0, 125 / 255.0)

custom_cmap = ListedColormap(
    [
        gray4,
        gray5,
        blue2,
        blue5,
        orange3,
        orange5,
        red5,
        red6,
        purple5,
        purple6,
        purple9,
    ]
)


class PlotScatter(PlotStructure):
    """
    PlotScatter handles plotting 2D scatter between a set of features.
    It will also optionally overlay KDE contours of the target variable,
    which is a first-order method for evaluating possible feature interactions
    and whether the learned relationships are consistent with the data.
    """

    oranges = ListedColormap(
        ["xkcd:peach", "xkcd:orange", "xkcd:bright orange", "xkcd:rust brown"]
    )
    blues = ListedColormap(
        ["xkcd:periwinkle blue", "xkcd:clear blue", "xkcd:navy blue"]
    )

    def __init__(self, BASE_FONT_SIZE=12):
        super().__init__(BASE_FONT_SIZE=BASE_FONT_SIZE)

    def plot_scatter(
        self,
        estimators,
        X,
        y,
        features,
        display_feature_names={},
        display_units={},
        subsample=1.0,
        peak_val=None,
        kde=False,
        **kwargs,
    ):
        """
        Plot KDE between two features and color-code by the target variable
        """

        # TODO: Plot relationships for multiple features!!

        estimator_names = list(estimators.keys())
        n_panels = len(estimator_names)
        only_one_model = len(estimator_names) == 1

        predictions = np.zeros((n_panels, len(y)), dtype=np.float16)
        j = 0
        for estimator_name, estimator in estimators.items():
            if hasattr(estimator, "predict_proba"):
                predictions[j, :] = estimator.predict_proba(X)[:, 1]
            else:
                predictions[j, :] = estimator.predict(X)
            j += 1

        kwargs = self.get_fig_props(n_panels, **kwargs)
        # create subplots, one for each feature
        fig, axes = self.create_subplots(
            n_panels=n_panels,
            sharex=False,
            sharey=False,
            **kwargs,
        )

        ax_iterator = self.axes_to_iterator(n_panels, axes)
        vmax = np.max(predictions)

        for i, ax in enumerate(ax_iterator):
            cf = self.scatter_(
                ax=ax,
                X=X,
                features=features,
                predictions=predictions[i, :],
                vmax=vmax,
                **kwargs,
            )

            if subsample < 1.0:
                size = min(int(subsample * len(y)), len(y))
                idxs = np.random.choice(len(y), size=size, replace=False)
                var1 = X[features[0]].values[idxs]
                var2 = X[features[1]].values[idxs]
                y1 = y.values[idxs]
            else:
                var1 = X[features[0]].values
                var2 = X[features[1]].values
                y1 = np.copy(y)

            # Shuffle values
            index = np.arange(len(var1))
            np.random.shuffle(index)
            var1 = var1[index]
            var2 = var2[index]
            y1 = y1[index]

            ax.set_xlabel(display_feature_names.get(features[0], features[0]))
            ax.set_ylabel(display_feature_names.get(features[1], features[1]))

            ax.grid(color="#2A3459", alpha=0.6, linestyle="dashed", linewidth=0.5)
            # bluish dark grey, but slightly lighter than background

            if kde:
                cmap_set = [
                    self.oranges,
                    self.blues,
                    "Reds",
                    "jet",
                ]
                classes = np.unique(y1)
                idx_sets = [np.where(y1 == c) for c in classes]
                for idxs, cmap in zip(idx_sets, cmap_set):
                    # Plot positive cases
                    cs = self.plot_kde_contours(
                        ax,
                        dy=var2[idxs],
                        dx=var1[idxs],
                        target=y1[idxs],
                        cmap=cmap,
                    )
                    handles_set = [cs.legend_elements()[-1]]
                labels = classes

                legend = ax.legend(handles_set, labels, framealpha=0.5)

            # Hide the right and top spines
            ax.spines["right"].set_visible(False)
            ax.spines["top"].set_visible(False)

            if not only_one_model:
                ax.set_title(estimator_names[i])

        if n_panels == 1:
            ax_ = axes
        else:
            ax_ = axes.ravel().tolist()

        if hasattr(estimator, "predict_proba"):
            cbar_label = "Probability"
        else:
            cbar_label = "Response"

        cbar_label = kwargs.get("cbar_label", cbar_label)

        fig.colorbar(cf, ax=ax_, label=cbar_label, orientation="horizontal")

        return fig, axes

    def scatter_(self, ax, features, X, predictions, vmax, **kwargs):
        """
        Plot 2D scatter of ML predictions;
        Only plots a random 20000 points
        """
        size = min(20000, len(X))
        idxs = np.random.choice(len(X), size=size, replace=False)

        x_val = X[features[0]].values[idxs]
        y_val = X[features[1]].values[idxs]
        z_val = predictions[idxs]

        # Show highest predictions on top!
        index = np.argsort(z_val)
        # index = np.random.choice(len(z_val), size=len(z_val), replace=False)

        x_val = x_val[index]
        y_val = y_val[index]
        z_val = z_val[index]

        cmap = kwargs.get("cmap", custom_cmap)
        zmax = vmax + 0.05 if vmax < 1.0 else 1.1
        delta = 0.05 if vmax < 0.5 else 0.1
        levels = [0, 0.05] + list(np.arange(0.1, zmax, delta))
        norm = mpl.colors.BoundaryNorm(levels, cmap.N)

        sca = ax.scatter(
            x_val,
            y_val,
            c=z_val,
            cmap=cmap,
            alpha=0.6,
            s=3,
            norm=norm,
        )
        return sca

    def kernal_density_estimate(self, dy, dx):
        dy_min = np.amin(dy)
        dx_min = np.amin(dx)
        dy_max = np.amax(dy)
        dx_max = np.amax(dx)

        x, y = np.mgrid[dx_min:dx_max:100j, dy_min:dy_max:100j]
        positions = np.vstack([x.ravel(), y.ravel()])
        values = np.vstack([dx, dy])
        kernel = gaussian_kde(values)
        f = np.reshape(kernel(positions).T, x.shape)

        return x, y, f

    def plot_kde_contours(
        self,
        ax,
        dy,
        dx,
        target,
        cmap,
    ):
        x, y, f = self.kernal_density_estimate(dy, dx)
        temp_linewidths = [0.85, 1.0, 1.25, 1.75]
        temp_thresh = [75.0, 90.0, 95.0, 97.5]
        temp_levels = [0.0, 0.0, 0.0, 0.0]
        for i in range(0, len(temp_thresh)):
            temp_levels[i] = np.percentile(f.ravel(), temp_thresh[i])

        # masked_f = np.ma.masked_where(f < 1.6e-5, f)
        cs = ax.contour(
            x,
            y,
            f,
            levels=temp_levels,
            cmap=cmap,
            linewidths=temp_linewidths,
            alpha=1.0,
        )
        fmt = {}
        for l, s in zip(cs.levels, temp_thresh[::-1]):
            fmt[l] = f"{int(s)}%"
        ax.clabel(
            cs, cs.levels, inline=True, fontsize=self.FONT_SIZES["teensie"], fmt=fmt
        )

        return cs



import numpy as np
import pandas as pd
from collections import OrderedDict
from .utils import is_list, to_list, is_fitted


class Attributes:
    """
    The Attributes class handles checking and setting the attributes
    for the InterpretToolkit, GlobalInterpret, and LocalInterpret classes.

    Attributes is a base class to be inherited by those classes and
    should never be instantiated
    """

    def set_estimator_attribute(self, estimator_objs, estimator_names):
        """
        Checks the type of the estimators and estimator_names attributes.
        If a list or not a dict, then the estimator argument
        is converted to a dict for processing.

        Parameters
        ----------
        estimators : object, list of objects
            A fitted estimator object or list thereof implementing `predict` or
            `predict_proba`.
            Multioutput-multiclass classifiers are not supported.

        estimator_names : string, list
            Names of the estimators (for internal and plotting purposes)
        """
        estimator_is_none = estimator_objs == None

        # Convert the estimator_objs to a list, if it is not already.
        if not is_list(estimator_objs):
            estimator_objs = to_list(estimator_objs)

        # Convert the name of the estimator_objs to a list,
        # if is not already.
        if not is_list(estimator_names):
            estimator_names = to_list(estimator_names)

        # Check that the estimator_objs and estimator_names are the same size.
        if not estimator_is_none:
            assert len(estimator_objs) == len(
                estimator_names
            ), "Number of estimator objects is not equal to the number of estimator names given!"

        # Check that the estimator objects have been fit!
        if not estimator_is_none:
            if not all([is_fitted(m) for m in estimator_objs]):
                raise ValueError(
                    "One or more of the estimators given has NOT been fit!"
                )

        # Create a dictionary from the estimator_objs and estimator_names.
        # Then set the attributes.
        self.estimators = OrderedDict(
            [(name, obj) for name, obj in zip(estimator_names, estimator_objs)]
        )
        self.estimator_names = estimator_names

    def set_y_attribute(self, y):
        """
        Checks the type of the y attribute.
        """
        # check that y are assigned correctly
        if type(y) == type(None):
            raise ValueError("y is required!")
        if is_list(y):
            self.y = np.array(y)
        elif isinstance(y, np.ndarray):
            self.y = y
        elif isinstance(y, (pd.DataFrame, pd.Series)):
            self.y = y.values
        else:
            if y is not None:
                raise TypeError("y must be an numpy array or pandas.DataFrame.")
            else:
                self.y = None

    def set_X_attribute(self, X, feature_names=None):
        """
        Check the type of the X attribute.
        """
        # make sure data is the form of a pandas dataframe regardless of input type
        if type(X) == type(None):
            raise ValueError("X are required!")
        if isinstance(X, np.ndarray):
            if feature_names is None:
                raise Exception("Feature names must be specified if using NumPy array.")
            else:
                self.X = pd.DataFrame(data=X, columns=feature_names)
        else:
            self.X = X

        if X is not None:
            self.feature_names = self.X.columns.to_list()

    def set_estimator_output(self, estimator_output, estimator):
        """
        Check the estimator output is given and if not try to
        assume the correct estimator output.
        """

        estimator_obj = estimator[0] if is_list(estimator) else estimator
        available_options = ["raw", "probability"]

        if estimator_output is None:
            if hasattr(estimator_obj, "predict_proba"):
                self.estimator_output = "probability"
            else:
                self.estimator_output = "raw"

        else:
            if estimator_output in available_options:
                self.estimator_output = estimator_output
            else:
                raise ValueError(
                    f"""
                                {estimator_output} is not an accepted options. 
                                 The available options are {available_options}.
                                 Check for syntax errors!
                                 """
                )



from __future__ import division

import numpy as np
import warnings

try:
    import matplotlib.pyplot as pl
    import matplotlib
except ImportError:
    warnings.warn("matplotlib could not be loaded!")
    pass

import seaborn as sns
from shap.plots import colors

try:
    from shap.utils import convert_name, approximate_interactions
except:
    from shap.common import convert_name, approximate_interactions

from matplotlib.ticker import MaxNLocator
from math import log10

from .base_plotting import PlotStructure

base_plot = PlotStructure()

labels = {
    "MAIN_EFFECT": "SHAP main effect value for\n%s",
    "INTERACTION_VALUE": "SHAP interaction value",
    "INTERACTION_EFFECT": "SHAP interaction value for\n%s and %s",
    "VALUE": "SHAP value (impact on model output)",
    "GLOBAL_VALUE": "mean(|SHAP value|) (average impact on model output magnitude)",
    "VALUE_FOR": "SHAP value for\n%s",
    "PLOT_FOR": "SHAP plot for %s",
    "FEATURE": "Feature %s",
    "FEATURE_VALUE": "Feature value",
    "FEATURE_VALUE_LOW": "Low",
    "FEATURE_VALUE_HIGH": "High",
    "JOINT_VALUE": "Joint SHAP value",
    "MODEL_OUTPUT": "Model output value",
}


def dependence_plot(
    feature,
    shap_values,
    X,
    feature_values=None,
    display_feature_names=None,
    interaction_index="auto",
    target_values=None,
    ax=None,
    fig=None,
    **kwargs
):
    """Create a SHAP dependence plot, colored by an interaction feature.

    Plots the value of the feature on the x-axis and the SHAP value of the same feature
    on the y-axis. This shows how the model depends on the given feature, and is like a
    richer extenstion of the classical parital dependence plots. Vertical dispersion of the
    data points represents interaction effects. Grey ticks along the y-axis are data
    points where the feature's value was NaN.


    Parameters
    ----------
    feature : string
        name of the feature to plot.

    shap_values : numpy.array
        Matrix of SHAP values (# samples x # features).

    X : pandas.DataFrame
        Matrix of feature values (# samples x # features).

    display_feature_names : list of strings
        A list of "prettier" names corresponding the column names
        in the X pandas.DataFrame

    interaction_index : "auto", None, int, or string
        The index of the feature used to color the plot. The name of a feature can also be passed
        as a string. If "auto" then shap.common.approximate_interactions is used to pick what
        seems to be the strongest interaction (note that to find to true stongest interaction you
        need to compute the SHAP interaction values).

    x_jitter : float (0 - 1)
        Adds random jitter to feature values. May increase plot readability when feature
        is discrete.

    alpha : float
        The transparency of the data points (between 0 and 1). This can be useful to the
        show density of the data points when using a large dataset.

    xmin : float or string
        Represents the lower bound of the plot's x-axis. It can be a string of the format
        "percentile(float)" to denote that percentile of the feature's value used on the x-axis.

    xmax : float or string
        Represents the upper bound of the plot's x-axis. It can be a string of the format
        "percentile(float)" to denote that percentile of the feature's value used on the x-axis.

    ax : matplotlib Axes object
         Optionally specify an existing matplotlib Axes object, into which the plot will be placed.
         In this case we do not create a Figure, otherwise we do.

    """
    marker = kwargs.get("marker", "o")
    unnormalize = kwargs.get("unnormalize", None)
    cmap = kwargs.get("cmap", colors.red_blue)
    feature_names = list(X.columns)

    color = kwargs.get("color", "#1E88E5")
    axis_color = kwargs.get("axis_color", "#333333")
    dot_size = kwargs.get("s", 5)
    x_jitter = kwargs.get("x_jitter", 0)
    xmin = kwargs.get("xmin", None)
    xmax = kwargs.get("xmax", None)
    alpha = kwargs.get("alpha", 0.8)

    histdata = kwargs.get("histdata", None)
    target = kwargs.get("target", None)

    if feature_values is None:
        feature_values = X.values

    # allow vectors to be passed
    if len(shap_values.shape) == 1:
        shap_values = np.reshape(shap_values, len(shap_values), 1)
    if len(X.shape) == 1:
        X = np.reshape(X, len(X), 1)

    feature_ind = convert_name(feature, shap_values, feature_names)

    # guess what other feature as the stongest interaction with the plotted feature
    interaction_index = get_interaction_index(
        feature_ind, interaction_index, shap_values, feature_values, feature_names
    )

    assert (
        shap_values.shape[0] == X.shape[0]
    ), "'shap_values' and 'features' values must have the same number of rows!"
    assert (
        shap_values.shape[1] == X.shape[1]
    ), "'shap_values' must have the same number of columns as 'features'!"

    # get both the raw and display feature values
    oinds = np.arange(shap_values.shape[0])

    # we randomize the ordering so plotting overlaps are not related to data ordering
    np.random.shuffle(oinds)

    xdata = feature_values[oinds, feature_ind].astype(np.float64)
    s = shap_values[oinds, feature_ind]
    if target_values is not None:
        target_values = target_values[oinds]

    # get both the raw and display color values
    if interaction_index is not None:
        cdata = feature_values[oinds, interaction_index]
        clow = np.nanpercentile(cdata.astype(np.float), 5)
        chigh = np.nanpercentile(cdata.astype(np.float), 95)
        if clow == chigh:
            clow = np.nanmin(cdata.astype(np.float))
            chigh = np.nanmax(cdata.astype(np.float))

    # optionally add jitter to feature values
    if x_jitter > 0:
        if x_jitter > 1:
            x_jitter = 1
        xvals = xdata.copy()
        if isinstance(xvals[0], float):
            xvals = xvals.astype(np.float)
            xvals = xvals[~np.isnan(xvals)]
        xvals = np.unique(xvals)  # returns a sorted array
        if len(xvals) >= 2:
            smallest_diff = np.min(np.diff(xvals))
            jitter_amount = x_jitter * smallest_diff
            xdata += (np.random.ranf(size=len(xdata)) * jitter_amount) - (
                jitter_amount / 2
            )

    # the actual scatter plot, TODO: adapt the dot_size to the number of data points?
    xdata_nan = np.isnan(xdata)
    xdata_notnan = np.invert(xdata_nan)

    ####
    if target_values is not None and interaction_index is not None:
        # Binary classification
        classes = np.unique(target_values)
        n_classes = len(classes)
        if n_classes == 2:
            dot_sizes = [1, 8]
            markers = ["o", "v"]
            idx_set = [np.where(target_values == i)[0] for i in classes]
            alphas = [0.4, 1.0]

    # Plot background histogram
    # add histogram
    if histdata is not None:
        hist_ax = make_twin_ax(ax)
        twin_yaxis_label = add_histogram_axis(
            hist_ax,
            feature=feature,
            target=target,
            data=histdata,
        )

    if interaction_index is not None and target_values is None:
        cdata_imp = cdata.copy()
        cdata_imp[np.isnan(cdata)] = (clow + chigh) / 2.0
        cdata[cdata_imp > chigh] = chigh
        cdata[cdata_imp < clow] = clow
        p = ax.scatter(
            xdata[xdata_notnan],
            s[xdata_notnan],
            s=dot_size,
            linewidth=0,
            c=cdata[xdata_notnan],
            cmap=cmap,
            alpha=alpha,
            vmin=clow,
            vmax=chigh,
            rasterized=len(xdata) > 500,
            label="SHAP",
            zorder=0,
        )
        p.set_array(cdata[xdata_notnan])

    elif target_values is not None and interaction_index is None:
        p = ax.scatter(
            xdata[xdata_notnan],
            s[xdata_notnan],
            s=dot_size,
            linewidth=0,
            c=target_values[xdata_notnan],
            cmap=cmap,
            alpha=alpha,
            vmin=np.min(target_values),
            vmax=np.max(target_values),
            rasterized=len(xdata) > 500,
            label="SHAP",
            zorder=0,
        )

    elif target_values is not None and interaction_index is not None:

        cdata_imp = cdata.copy()
        cdata_imp[np.isnan(cdata)] = (clow + chigh) / 2.0
        cdata[cdata_imp > chigh] = chigh
        cdata[cdata_imp < clow] = clow
        for idxs, dot_size, marker, alpha in zip(idx_set, dot_sizes, markers, alphas):
            xdata_temp = xdata[xdata_notnan]
            ydata_temp = s[xdata_notnan]
            cdata_temp = cdata[xdata_notnan]

            p = ax.scatter(
                xdata_temp[idxs],
                ydata_temp[idxs],
                s=dot_size,
                linewidth=0,
                c=cdata_temp[idxs],
                cmap=cmap,
                alpha=alpha,
                vmin=clow,
                vmax=chigh,
                rasterized=len(xdata) > 500,
                marker=marker,
                label="SHAP",
                zorder=0,
            )

    else:
        p = ax.scatter(
            xdata[xdata_notnan],
            s[xdata_notnan],
            s=dot_size,
            linewidth=0,
            color=color,
            alpha=alpha,
            rasterized=len(xdata) > 500,
            marker=marker,
            label="SHAP",
            zorder=0,
        )

    if interaction_index != feature_ind and interaction_index is not None:
        # draw the color bar
        pad = 0.05 if histdata is None else 0.18
        pad = kwargs.get('colorbar_pad', pad) 
        cb = pl.colorbar(p, ticks=MaxNLocator(5), ax=ax, pad=pad)
        cb.set_label(display_feature_names[interaction_index], size=8)
        cb.ax.tick_params(labelsize=8)
        cb.set_alpha(1)
        cb.outline.set_visible(False)
        bbox = cb.ax.get_window_extent().transformed(fig.dpi_scale_trans.inverted())
        cb.ax.set_aspect((bbox.height - 0.7) * 20)
        base_plot._to_sci_notation(ax=None, colorbar=cb, ydata=cdata)

    # plot any nan feature values as tick marks along the y-axis
    # NOT plotting the full range.
    xmin = np.nanpercentile(xdata, 2.0)
    xmax = np.nanpercentile(xdata, 97.5)
    ax.set_xlim([xmin, xmax])

    xlim = ax.get_xlim()
    if interaction_index is not None:
        p = ax.scatter(
            xlim[0] * np.ones(xdata_nan.sum()),
            s[xdata_nan],
            marker=1,
            linewidth=2,
            c=cdata_imp[xdata_nan],
            cmap=cmap,
            alpha=alpha,
            vmin=clow,
            vmax=chigh,
            rasterized=len(xdata) > 1000,
            zorder=0,
        )
        p.set_array(cdata[xdata_nan])
    else:
        ax.scatter(
            xlim[0] * np.ones(xdata_nan.sum()),
            s[xdata_nan],
            marker=1,
            linewidth=2,
            color=color,
            alpha=alpha,
            rasterized=len(xdata) > 1000,
            zorder=0,
        )

    # NOT plotting the full range.
    # xmin = np.nanpercentile(xdata, 1.0)
    # xmax = np.nanpercentile(xdata, 99.0)
    # ax.set_xlim([xmin, xmax])

    # make the plot more readable
    ax.xaxis.set_ticks_position("bottom")
    ax.yaxis.set_ticks_position("left")
    ax.spines["right"].set_visible(False)
    ax.spines["top"].set_visible(False)
    ax.tick_params(color=axis_color, labelcolor=axis_color, labelsize=11)
    for spine in ax.spines.values():
        spine.set_edgecolor(axis_color)


def get_interaction_index(
    feature_ind, interaction_index, shap_values, feature_values, feature_names
):
    """ """
    # guess what other feature as the stongest interaction with the plotted feature
    if not hasattr(feature_ind, "__len__"):
        if interaction_index == "auto":
            interaction_index = approximate_interactions(
                feature_ind, shap_values, feature_values
            )[0]
        interaction_index = convert_name(interaction_index, shap_values, feature_names)

    return interaction_index


def add_histogram_axis(
    ax,
    data,
    feature,
    target,
    bins="auto",
    min_value=None,
    max_value=None,
    density=False,
    **kwargs
):
    """
    Adds a background histogram of data for a given feature.
    """
    sns.histplot(
        ax=ax,
        data=data,
        x=feature,
        hue=target,
        legend=False,
        stat="probability",
        common_norm=False,
    )
    ax.set_ylabel("")

    """
        color = kwargs.get("color", "xkcd:dark sky blue")
        edgecolor = kwargs.get("color", "white")

        cnt, bins, patches = ax.hist(
            data,
            bins=bins,
            alpha=0.3,
            color=color,
            density=density,
            edgecolor=edgecolor,
            zorder=1,
        )

        if density:
            return "Relative Frequency"
        else:
            ax.set_yscale("log")
            ymax = round(10 * len(data))
            n_ticks = round(log10(ymax))
            ax.set_ylim([0, ymax])
            ax.set_yticks([10 ** i for i in range(n_ticks + 1)])
            return "Frequency"
        """


def make_twin_ax(ax):
    """
    Create a twin axis on an existing axis with a shared x-axis
    """
    # align the twinx axis
    twin_ax = ax.twinx()

    # Turn twin_ax grid off.
    twin_ax.grid(False)

    # Set ax's patch invisible
    ax.patch.set_visible(False)
    # Set axtwin's patch visible and colorize it in grey
    twin_ax.patch.set_visible(True)

    # move ax in front
    ax.set_zorder(twin_ax.get_zorder() + 1)

    return twin_ax



from .base_plotting import PlotStructure
from ..common.utils import to_list, is_list

import numpy as np
import matplotlib.pyplot as plt
from matplotlib.colors import BoundaryNorm
from matplotlib import colors
import itertools
import scipy.stats as sps
import numpy as np
from matplotlib.patches import Rectangle
from mpl_toolkits.axes_grid1.inset_locator import inset_axes


class MidpointNormalize(colors.Normalize):
    def __init__(self, vmin=None, vmax=None, vcenter=None, clip=False):
        self.vcenter = vcenter
        colors.Normalize.__init__(self, vmin, vmax, clip)

    def __call__(self, value, clip=None):
        # I'm ignoring masked values and all kinds of edge cases to make a
        # simple example...
        x, y = [self.vmin, self.vcenter, self.vmax], [0, 0.5, 1]
        return np.ma.masked_array(np.interp(value, x, y))


class PlotInterpret2D(PlotStructure):
    """
    PlotInterpret2D handles the 2D explainability graphics, which include
    the 2D ALE and PD.
    """

    def __init__(self, BASE_FONT_SIZE=12):
        super().__init__(BASE_FONT_SIZE=BASE_FONT_SIZE)

    def add_histogram_axis(
        self,
        ax,
        data,
        bins=15,
        min_value=None,
        max_value=None,
        density=True,
        orientation="vertical",
        **kwargs,
    ):
        """
        Adds a background histogram of data for a given feature.
        """
        color = kwargs.get("color", "xkcd:steel")
        edgecolor = kwargs.get("color", "white")

        # data = np.clip(data, a_min=np.nan, a_max=np.nan)

        cnt, bins, patches = ax.hist(
            data,
            bins=bins,
            alpha=0.35,
            color=color,
            density=density,
            edgecolor=edgecolor,
            orientation=orientation,
            zorder=1,
        )

        # data = np.sort(np.random.choice(data, size=10000), replace=True)
        # kde = sps.gaussian_kde(data)
        # kde_pdf = kde.pdf(data)

        # if orientation == 'vertical':
        # ax.plot(data, kde_pdf, linewidth=0.5, color='xkcd:darkish blue', alpha=0.9)
        # ax.set_ylim([0, 1.75*np.amax(kde_pdf)])
        # else:
        # ax.plot(kde_pdf, data, linewidth=0.5, color='xkcd:darkish blue', alpha=0.9)
        # ax.set_xlim([0,1.75*np.amax(kde_pdf)])

    def plot_2d_kde(self, ax, x, y):
        """
        Add contours of the kernel density estimate
        """
        xmin, xmax = np.min(x), np.max(x)
        ymin, ymax = np.min(y), np.max(y)

        # Peform the kernel density estimate
        xx, yy = np.mgrid[xmin:xmax:100j, ymin:ymax:100j]
        positions = np.vstack([xx.ravel(), yy.ravel()])
        values = np.vstack([x, y])
        kernel = sps.gaussian_kde(values)
        f = np.reshape(kernel(positions).T, xx.shape)

        percentiles = [50.0, 75.0, 90.0]
        linewidths = [
            0.25,
            0.75,
            1.5,
        ]
        line_colors = ["xkcd:indigo", "xkcd:dark cyan", "xkcd:dandelion"]
        levels = np.zeros((len(percentiles)))
        for i in range(len(percentiles)):
            levels[i] = np.percentile(f.ravel(), percentiles[i])

        # Contour plot
        cset = ax.contour(
            xx,
            yy,
            f,
            levels=levels,
            linewidths=linewidths,
            colors=line_colors,
        )
        fmt = {}
        for l, s in zip(cset.levels, percentiles[::-1]):
            fmt[l] = f"{int(s)}"

        ax.clabel(cset, cset.levels, inline=True, fontsize=6, fmt=fmt)

    def plot_contours(
        self,
        method,
        data,
        features,
        estimator_names,
        display_feature_names={},
        display_units={},
        to_probability=False,
        **kwargs,
    ):

        """
        Generic function for 2-D PDP/ALE
        """
        contours = kwargs.get("contours", False)
        kde_curves = kwargs.get("kde_curves", True)
        scatter = kwargs.get("scatter", True)

        if not is_list(estimator_names):
            estimator_names = to_list(estimator_names)
   
        unnormalize = kwargs.get("unnormalize", None)
        self.display_feature_names = display_feature_names
        self.display_units = display_units

        if not isinstance(features, list):
            features = [features]

        hspace = kwargs.get("hspace", 0.4)
        wspace = kwargs.get("wspace", 0.7)
        cmap = kwargs.get("cmap", "seismic")
        colorbar_label = kwargs.get("left_yaxis_label")

        if colorbar_label == "Centered ALE (%)":
            colorbar_label = "2nd Order ALE (%)"

        # get the number of panels which will be length of feature dictionary
        n_panels = len(features) * len(estimator_names)
        n_columns = len(estimator_names)

        if len(estimator_names) == 1:
            only_one_model = True
            n_columns = min(len(features), 3)
        else:
            only_one_model = False
            n_columns = len(estimator_names)

        if n_panels == 1:
            figsize = (6, 3)
            fontsize = 8
        else:
            figsize = (11, 6)
            fontsize = 8

        # create subplots, one for each feature
        fig, main_axes, top_axes, rhs_axes, n_rows = self._create_joint_subplots(
            n_panels=n_panels, n_columns=n_columns, figsize=figsize, ratio=5
        )

        is_even = (n_rows * n_columns) / (n_panels)
        feature_levels = {f: {"max": [], "min": []} for f in features}

        ale_max = []
        ale_min = []
        for feature_set, model_name in itertools.product(features, estimator_names):
            zdata_temp = data[
                f"{feature_set[0]}__{feature_set[1]}__{model_name}__{method}"
            ].values.copy()
            zdata_temp = np.ma.getdata(zdata_temp)

            if to_probability:
                zdata_temp *= 100.0

            # if the max value is super small
            if np.round(np.max(np.absolute(zdata_temp)), 10) < 1e-5:
                zdata_temp[:] = 0.0

            max_value = np.nanmax(zdata_temp)
            min_value = np.nanmin(zdata_temp)

            if np.all((zdata_temp == 0.0)):
                ale_max.append(0.01)
                ale_min.append(-0.01)
                feature_levels[feature_set]["max"].append(0.01)
                feature_levels[feature_set]["min"].append(-0.01)
            else:
                ale_max.append(max_value)
                ale_min.append(min_value)
                feature_levels[feature_set]["max"].append(max_value)
                feature_levels[feature_set]["min"].append(min_value)

        levels = self.calculate_ticks(
            nticks=20,
            upperbound=np.nanpercentile(ale_max, 100),
            lowerbound=np.nanpercentile(ale_min, 0),
            round_to=5,
            center=True,
        )

        cmap = plt.get_cmap(cmap)
        counter = 0
        n = 1
        i = 0
        for feature_set, model_name in itertools.product(features, estimator_names):
            # We want to lowest maximum value and the highest minimum value
            if not only_one_model:
                max_value = np.nanpercentile(feature_levels[feature_set]["max"], 100)
                min_value = np.nanpercentile(feature_levels[feature_set]["min"], 0)
                levels = self.calculate_ticks(
                    nticks=20,
                    upperbound=max_value,
                    lowerbound=min_value,
                    round_to=5,
                    center=True,
                )

            main_ax = main_axes[i]
            top_ax = top_axes[i]
            rhs_ax = rhs_axes[i]

            if counter <= len(estimator_names) - 1 and not only_one_model:
                top_ax.set_title(
                    model_name, fontsize=self.FONT_SIZES["normal"], alpha=0.9
                )
                counter += 1

            xdata1 = data[f"{feature_set[0]}__bin_values"].values
            xdata2 = data[f"{feature_set[1]}__bin_values"].values

            xdata1_hist = data[f"{feature_set[0]}"].values
            xdata2_hist = data[f"{feature_set[1]}"].values
            if (unnormalize is not None) and (unnormalize[model_name] is not None):
                unnorm_obj = unnormalize[model_name]

                xdata1_hist = unnorm_obj.inverse_transform(xdata1_hist, feature_set[0])
                xdata2_hist = unnorm_obj.inverse_transform(xdata2_hist, feature_set[1])
                xdata1 = unnorm_obj.inverse_transform(xdata1, feature_set[0])
                xdata2 = unnorm_obj.inverse_transform(xdata2, feature_set[1])

            zdata = data[
                f"{feature_set[0]}__{feature_set[1]}__{model_name}__{method}"
            ].values.copy()

            # can only do a contour plot with 2-d data
            x1, x2 = np.meshgrid(xdata1, xdata2, indexing="xy")

            # Get the mean of the bootstrapping.
            if np.ndim(zdata) > 2:
                zdata = np.mean(zdata, axis=0)
            else:
                zdata = zdata.squeeze()

            if to_probability:
                zdata *= 100.0

            if contours:
                cf = main_ax.contourf(
                    x1, x2, zdata, cmap=cmap, alpha=0.8, levels=levels, extend="neither"
                )
            else:
                cf = main_ax.pcolormesh(
                    x1,
                    x2,
                    zdata,
                    cmap=cmap,
                    alpha=0.8,
                    norm=BoundaryNorm(boundaries=levels, ncolors=cmap.N, clip=True),
                )
            
            """
            mark_empty = False
            if mark_empty:
                # Do not autoscale, so that boxes at the edges (contourf only plots the bin
                # centres, not their edges) don't enlarge the plot.
                plt.autoscale(False)
                # Add rectangles to indicate cells without samples.
                for i, j in zip(*np.where(masked)):
                    main_ax.add_patch(
                        Rectangle(
                            [xdata1[i], xdata2[j]],
                            xdata1[i + 1] - xdata1[i],
                            xdata2[j + 1] - xdata2[j],
                            linewidth=1,
                            edgecolor="k",
                            facecolor="none",
                            alpha=0.4,
                        )
                    )
            """

            if scatter:
                idx = np.random.choice(
                    len(xdata1_hist), size=min(2000, len(xdata1_hist))
                )
                main_ax.scatter(
                    xdata1_hist[idx], xdata2_hist[idx], alpha=0.3, color="grey", s=1
                )

            if kde_curves:
                try:
                    # There can be very rare cases where two functions are linearly correlated (cc~1.0)
                    # which can cause the KDE calculations to fail!
                    self.plot_2d_kde(main_ax, xdata1_hist, xdata2_hist)
                except:
                    pass

            self.add_histogram_axis(
                top_ax,
                xdata1_hist,
                bins=30,
                orientation="vertical",
                min_value=xdata1[1],
                max_value=xdata1[-2],
            )

            self.add_histogram_axis(
                rhs_ax,
                xdata2_hist,
                bins=30,
                orientation="horizontal",
                min_value=xdata2[1],
                max_value=xdata2[-2],
            )

            main_ax.set_ylim([xdata2[0], xdata2[-1]])
            main_ax.set_xlim([xdata1[0], xdata1[-1]])

            self.set_minor_ticks(main_ax)
            self.set_axis_label(
                main_ax,
                xaxis_label=feature_set[0],
                yaxis_label=feature_set[1],
                fontsize=fontsize,
            )
            # Add a colorbar
            if (
                i == (n * n_columns - 1) or (i == len(main_axes) - 1 and is_even > 1)
            ) and not only_one_model:
                self.add_colorbar(
                    fig,
                    plot_obj=cf,
                    ax=rhs_ax,
                    colorbar_label=colorbar_label,
                    extend="both",
                )
                n += 1
            i += 1

        if only_one_model:
            major_ax = self.set_major_axis_labels(fig=fig)
            # colorbar
            #cax = major_ax.inset_axes(
            #    major_ax,
            #    width="100%",  # width = 10% of parent_bbox width
            #    height="100%",  # height : 50%
            #    loc="lower center",
            #    bbox_to_anchor=(0.02, -0.1, 0.8, 0.05),
            #    bbox_transform=major_ax.transAxes,
            #    borderpad=0,
            #)
            
            cax = major_ax.inset_axes(
                bounds = (0.02, -0.1, 0.8, 0.05),
                transform=major_ax.transAxes,
            )
            #    width="100%",  # width = 10% of parent_bbox width
            #    height="100%",  # height : 50%
            #    loc="lower center",
            #    borderpad=0,
            #)
            
            
            self.add_colorbar(
                fig,
                plot_obj=cf,
                cax=cax,
                orientation="horizontal",
                pad=0,
                shrink=0.8,
                colorbar_label=colorbar_label,
                extend="both",
            )
            
        # Add an letter per panel for publication purposes.
        self.add_alphabet_label(n_panels, main_axes)

        fig.tight_layout()
        fig.subplots_adjust(hspace=0.1, wspace=0.05)

        return fig, main_axes



"""These utilities are designed to check whether the given data and variable
names match the expected format. For the training or scoring data, we accept 
either a pandas dataframe with the target column indicated, two different 
dataframes, or two numpy arrays"""

import numpy as np
import pandas as pd

from .error_handling import InvalidDataException, InvalidInputException

try:
    basestring
except NameError:  # Python3
    basestring = str

__all__ = ["verify_data", "determine_variable_names"]


def verify_data(data):
    """Verifies that the data tuple is of the right format and coerces it to
    numpy arrays for the code under the hood

    :param data: one of the following:
        (pandas dataframe, string for target column),
        (pandas dataframe for inputs, pandas dataframe for outputs),
        (numpy array for inputs, numpy array for outputs)
    :returns: (numpy array for input, numpy array for output) or
        (pandas dataframe for input, pandas dataframe for output)
    """
    try:
        iter(data)
    except TypeError:
        raise InvalidDataException(data, "Data must be iterable")
    else:
        if len(data) != 2:
            raise InvalidDataException(data, "Data must contain 2 elements")
        else:
            # check if the first element is pandas dataframe or numpy array
            if isinstance(data[0], pd.DataFrame):
                # check if the second element is string or pandas dataframe
                if isinstance(data[1], basestring):
                    return (
                        data[0].loc[:, data[0].columns != data[1]],
                        data[0][[data[1]]],
                    )
                elif isinstance(data[1], pd.DataFrame):
                    return data[0], data[1]
                else:
                    raise InvalidDataException(
                        data,
                        "Second element of data must be a string for the target column or a pandas dataframe",
                    )

            elif isinstance(data[0], np.ndarray):
                if isinstance(data[1], np.ndarray):
                    return data[0], data[1]
                else:
                    raise InvalidDataException(
                        data, "Second element of data must also be a numpy array"
                    )
            else:
                raise InvalidDataException(
                    data,
                    "First element of data must be a numpy array or pandas dataframe",
                )


def determine_variable_names(data, variable_names):
    """Uses ``data`` and/or the ``variable_names`` to determine what the
    variable names are. If ``variable_names`` is not specified and ``data`` is
    not a pandas dataframe, defaults to the column indices

    :param data: a 2-tuple where the input data is the first item
    :param variable_names: either a list of variable names or None
    :returns: a list of variable names
    """
    if variable_names is not None:
        try:
            iter(variable_names)
        except TypeError:
            raise InvalidInputException(
                variable_names, "Variable names must be iterable"
            )
        else:
            if len(variable_names) != data[0].shape[1]:
                raise InvalidInputException(
                    variable_names,
                    "Variable names should have length %i" % data[0].shape[1],
                )
            else:
                return np.array(variable_names)
    else:
        if isinstance(data[0], pd.DataFrame):
            return data[0].columns.values
        else:
            return np.arange(data[0].shape[1])



"""There are a handful of different errors and warnings that we can report. This
houses all of them and provides information regarding ways to fix them."""


class InvalidStrategyException(Exception):
    """Thrown when a scoring strategy is invalid"""

    def __init__(self, strategy, msg=None, options=None):
        if msg is None:
            msg = (
                "%s is not a valid strategy for determining the optimal variable. "
                % strategy
            )
            msg += "\nShould be a callable or a valid string option. "
            if options is not None:
                msg += "Valid options are\n%r" % options

        super(InvalidStrategyException, self).__init__(msg)
        self.strategy = strategy
        self.options = None


class InvalidInputException(Exception):
    """Thrown when the input to the program does not match expectations"""

    def __init__(self, value, msg=None):
        if msg is None:
            msg = "Input value does not match expectations: %s" % value

        super(InvalidInputException, self).__init__(msg)
        self.value = value


class InvalidDataException(Exception):
    """Thrown when the training or scoring data is not of the right type"""

    def __init__(self, data, msg=None):
        if msg is None:
            msg = "Data is not of the right format"

        super(InvalidDataException, self).__init__(msg)
        self.data = data


class UnmatchedLengthPredictionsException(Exception):
    """Thrown when the number of predictions doesn't match truths"""

    def __init__(self, truths, predictions, msg=None):
        if msg is None:
            msg = "Shapes of truths and predictions do not match: %r and %r" % (
                truths.shape,
                predictions.shape,
            )

        super(UnmatchedLengthPredictionsException, self).__init__(msg)
        self.truths = truths
        self.predictions = predictions


class UnmatchingProbabilisticForecastsException(Exception):
    """Thrown when the shape of probabilisic predictions doesn't match the truths"""

    def __init__(self, truths, predictions, msg=None):
        if msg is None:
            msg = "Shapes of truths and predictions do not match: %r and %r" % (
                truths.shape,
                predictions.shape,
            )

        super(UnmatchingProbabilisticForecastsException, self).__init__(msg)
        self.truths = truths
        self.predictions = predictions


class AmbiguousProbabilisticForecastsException(Exception):
    """Thrown when classes were not provided for converting probabilistic
    predictions to deterministic ones but are required"""

    def __init__(self, truths, predictions, msg=None):
        if msg is None:
            msg = "Classes not provided for converting probabilistic predictions to deterministic ones"

        super(AmbiguousProbabilisticForecastsException, self).__init__(msg)
        self.truths = truths
        self.predictions = predictions


class FullImportanceResultWarning(Warning):
    """Thrown when we try to add a result to a full
    :class:`PermutationImportance.result.ImportanceResult`"""

    pass
